/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution. */

/* iem-gui-lib written by Thomas Musil (c) IEM KUG Graz Austria 2000-2001 */

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#ifndef  PD_VERSION_MINOR
#define PD_VERSION_MINOR 32
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "m_imp.h"
#include "g_canvas.h"
#include "t_tk.h"
#include "g_7_guis.h"

#ifdef NT
#include <io.h>
#else
#include <unistd.h>
#endif

t_symbol *iem_gui_key_sym=0;

int iem_gui_color_hex[]=
{
    16579836, 10526880, 4210752, 16572640, 16572608,
    16579784, 14220504, 14220540, 14476540, 16308476,
    14737632, 8158332, 2105376, 16525352, 16559172,
    15263784, 1370132, 2684148, 3952892, 16003312,
    12369084, 6316128, 0, 9177096, 5779456,
    7874580, 2641940, 17488, 5256, 5767248
};

int iem_gui_clip_size(int size)
{
    if(size < IEM_GUI_MINSIZE)
	size = IEM_GUI_MINSIZE;
    return(size);
}

int iem_gui_clip_font(int size)
{
    if(size < IEM_FONT_MINSIZE)
	size = IEM_FONT_MINSIZE;
    return(size);
}

int iem_gui_modulo_color(int col)
{
    while(col >= IEM_GUI_MAX_COLOR)
	col -= IEM_GUI_MAX_COLOR;
    while(col < 0)
	col += IEM_GUI_MAX_COLOR;
    return(col);
}

void str_xch_raute2dollar(t_symbol *s)
{
    if(s->s_name[0] == '#')
	s->s_name[0] = '$';
}

void str_xch_dollar2raute(t_symbol *s)
{
    if(s->s_name[0] == '$')
	s->s_name[0] = '#';
}

void reconvert2unique(t_symbol *s, int andunique)
{
    if(andunique && (s->s_name[0] != '$'))
    {
	int i, j=2;

        s->s_name[0] = '$';
	for(i=1; i<6; i++)
	{
	    if(((int)(s->s_name[i] - '0') < 0) || ((int)(s->s_name[i] - '0') > 9))
	    {
		j = i;
                i = 200;
	    }
	}
        s->s_name[1] = '0';
        for(i=2; i<100; i++, j++)
	{
	    if(s->s_name[j] == 0)
	    {
                s->s_name[i] = 0;
		i = 200;
	    }
	    else
	    {
                s->s_name[i] = s->s_name[j];
	    }
	}
    }
}

int sym_is_dollarunique(t_symbol *s)
{
    if((s->s_name[0] == '$') && (s->s_name[1] == '0'))
	return(1);
    else
        return(0);
}

/* --------------- bng     gui-bang ------------------------- */

t_widgetbehavior bng_widgetbehavior;
static t_class *bng_class;

typedef struct _bng
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_size;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     int      x_flashed;
     int      x_fcol;
     int      x_bcol;
     int      x_lcol;
     int      x_unique;
     int      x_loadinit;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_selected;
     int      x_shiftdown;
     int      x_flashtime_break;
     int      x_flashtime_hold;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
     t_clock *x_clock_hld;
     t_clock *x_clock_brk;
} t_bng;

/* widget helper functions */

void bng_update(t_bng *x, t_glist *glist)
{
    if(glist_isvisible(glist))
    {
	sys_vgui(".x%x.c itemconfigure %xBUT -fill #%6.6x\n", glist_getcanvas(glist), x,
		 x->x_flashed?x->x_fcol:x->x_bcol);
    }
}

void bng_draw(t_bng *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(firsttime)/* bang-rect + bang-circ */
    {
	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size,
		 x->x_bcol, x);
	sys_vgui(".x%x.c create oval %d %d %d %d -fill #%6.6x -tags %xBUT\n",
		 canvas, x->x_obj.te_xpos+1, x->x_obj.te_ypos+1,
		 x->x_obj.te_xpos + x->x_size-1, x->x_obj.te_ypos + x->x_size-1,
		 x->x_flashed?x->x_fcol:x->x_bcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize, x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize, x->x_lcol, x);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas, x->x_obj.te_xpos,
		 x->x_obj.te_ypos + x->x_size-1, x->x_obj.te_xpos + IOWIDTH,
		 x->x_obj.te_ypos + x->x_size, x, 0);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + IOWIDTH, x->x_obj.te_ypos+1, x, 0);
    }
    else
    {
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size);
	sys_vgui(".x%x.c coords %xBUT %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos+1,x->x_obj.te_ypos+1,
		 x->x_obj.te_xpos + x->x_size-1, x->x_obj.te_ypos + x->x_size-1);
        sys_vgui(".x%x.c itemconfigure %xBUT -fill #%6.6x\n", canvas, x,
		 x->x_flashed?x->x_fcol:x->x_bcol);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
        sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0, x->x_obj.te_xpos,
		 x->x_obj.te_ypos + x->x_size-1, x->x_obj.te_xpos + IOWIDTH,
		 x->x_obj.te_ypos + x->x_size);
        sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + IOWIDTH, x->x_obj.te_ypos+1);
    }
}

void bng_erase(t_bng* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xBUT\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void bng_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_bng *x = (t_bng *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void bng_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_bng *x = (t_bng *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    bng_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void bng_select(t_gobj *z, t_glist *glist, int selected)
{
    t_bng *x = (t_bng *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
	sys_vgui(".x%x.c itemconfigure %xBUT -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
	sys_vgui(".x%x.c itemconfigure %xBUT -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void bng_delete(t_gobj *z, t_glist *glist)
{
    t_bng *x = (t_bng *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void bng_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_bng *x = (t_bng *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 bng_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 bng_erase(x, glist);
         rtext_free(y);
    }
}

static void bng_save(t_gobj *z, t_binbuf *b)
{
    t_bng *x = (t_bng *)z;
    int bcol, fcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    fcol = -1 - (((0xfc0000 & x->x_fcol) >> 6)|((0xfc00 & x->x_fcol) >> 4)|((0xfc & x->x_fcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiiisssiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("bng"), x->x_size,
		x->x_flashtime_hold, x->x_flashtime_break,
		x->x_loadinit,
                x->x_snd, x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                ((int)(x->x_font[0]-'a'+1)/8)+x->x_unique, x->x_fontsize,
		bcol, fcol, lcol);
    binbuf_addv(b, ";");
}

void bng_check_minmax(t_bng *x, int ftbreak, int fthold)
{
    if(ftbreak > fthold)
    {
	int h;

	h = ftbreak;
	ftbreak = fthold;
	fthold = h;
    }
    if(ftbreak < IEM_BNG_MINBREAKFLASHTIME)
	ftbreak = IEM_BNG_MINBREAKFLASHTIME;
    if(fthold < IEM_BNG_MINHOLDFLASHTIME)
        fthold = IEM_BNG_MINHOLDFLASHTIME;
    x->x_flashtime_break = ftbreak;
    x->x_flashtime_hold = fthold;
}

static void bng_properties(t_gobj *z, t_glist *owner)
{
    t_bng *x = (t_bng *)z;
    char buf[600];

    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);

    sprintf(buf, "pdtk_iemgui_dialog %%s BANG \
	    ----------dimensions(pix):----------- %g %g size: 0 0 empty \
	    --------flash-time(ms)(ms):--------- %g intrrpt: %g hold: %g \
	    %g empty empty %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_size, (float)IEM_GUI_MINSIZE,
	    (float)x->x_flashtime_break, (float)x->x_flashtime_hold, 2.0,/*min_max_schedule+clip*/
	    -1.0, (float)x->x_loadinit, -1.0,/*no linlog, no multi*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, 0xffffff & x->x_fcol, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);

    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void bng_bang(t_bng *x)
{
    if(x->x_flashed)
    {
	x->x_flashed = 0;
	bng_update(x, x->x_glist);
	clock_delay(x->x_clock_brk, x->x_flashtime_break);
        x->x_flashed = 1;
    }
    else
    {
	x->x_flashed = 1;
	bng_update(x, x->x_glist);
    }
    clock_delay(x->x_clock_hld, x->x_flashtime_hold);
    outlet_bang(x->x_obj.ob_outlet);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_bang(x->x_snd->s_thing);
}

static void bng_dialog(t_bng *x, t_symbol *s, int argc, t_atom *argv)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int a = (int)atom_getintarg(0, argc, argv);
    int fthold = (int)atom_getintarg(2, argc, argv);
    int ftbreak = (int)atom_getintarg(3, argc, argv);
    int init = (int)atom_getintarg(5, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int fcol = (int)atom_getintarg(15, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    if(init != 0) init = 1;
    x->x_loadinit = init;

    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(snd))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;

    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }

    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_size = iem_gui_clip_size(a);
    bng_check_minmax(x, ftbreak, fthold);
    x->x_lcol = lcol & 0xffffff;
    x->x_fcol = fcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
       fs = 4;
    x->x_fontsize = fs;

    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, x->x_lab->s_name);
    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x, x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xBUT -fill #%6.6x\n", canvas, x,
	         x->x_flashed?x->x_fcol:x->x_bcol);
    bng_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(canvas, (t_text*)x);
}

static void bng_click(t_bng *x, t_floatarg xpos, t_floatarg ypos, t_floatarg shift, t_floatarg ctrl, t_floatarg alt)
{
    bng_bang(x);
}

#if (PD_VERSION_MINOR > 31)
static int bng_newclick(t_gobj *z, struct _glist *glist, int xpix, int ypix, int shift, int alt, int dbl, int doit)
{
    if(doit)
	bng_click((t_bng *)z, (t_floatarg)xpix, (t_floatarg)ypix, (t_floatarg)shift, 0, (t_floatarg)alt);
    return (1);
}
#endif

static void bng_float(t_bng *x, t_floatarg f)
{
    bng_bang(x);
}

static void bng_symbol(t_bng *x, t_symbol *s)
{
    bng_bang(x);
}

static void bng_pointer(t_bng *x, t_gpointer *gp)
{
    bng_bang(x);
}

static void bng_anything(t_bng *x, t_symbol *s, int argc, t_atom *argv)
{
    bng_bang(x);
}

static void bng_loadbang(t_bng *x)
{
    if(!sys_noloadbang && x->x_loadinit)
	bng_bang(x);
}

static void bng_size(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_size = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    if(glist_isvisible(x->x_glist))
    { 
        bng_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void bng_delta(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        bng_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void bng_pos(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        bng_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void bng_flashtime(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    bng_check_minmax(x, (int)atom_getintarg(0, ac, av),
			 (int)atom_getintarg(1, ac, av));
}

static void bng_color(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i,j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_fcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_fcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(2, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas,
		 x, x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
	sys_vgui(".x%x.c itemconfigure %xBUT -fill #%6.6x\n", canvas,
		 x, x->x_flashed?x->x_fcol:x->x_bcol);
	sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
		 x, x->x_bcol);

    }
}

static void bng_send(t_bng *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;

    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void bng_receive(t_bng *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;

    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void bng_label(t_bng *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;

    x->x_lab = s;


    if(glist_isvisible(x->x_glist))
    {
	if(!strcmp(s->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, s->s_name);
    }
}

static void bng_label_pos(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 glist_getcanvas(x->x_glist), x,
	     x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
    }
}

static void bng_label_font(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
       f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
    {
        sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
    }
}

static void bng_init(t_bng *x, t_floatarg f)
{
    x->x_loadinit = (f==0.0)?0:1;
}

static void bng_list(t_bng *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac == 2)&&IS_A_FLOAT(av,0)&&IS_A_SYMBOL(av,1))
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;

		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
		    {
			bng_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
	else
	    bng_bang(x);
    }
    else
	bng_bang(x);
}

static void bng_tick_hld(t_bng *x)
{
    x->x_flashed = 0;
    bng_update(x, x->x_glist);
}

static void bng_tick_brk(t_bng *x)
{
    bng_update(x, x->x_glist);
}

static void *bng_new(t_symbol *s, int argc, t_atom *argv)
{
    t_bng *x = (t_bng *)pd_new(bng_class);
    int bcol= -262144;
    int fcol= -1;
    int lcol= -1, a=IEM_GUI_DEFAULTSIZE;
    int init=0, ldx=20, ldy=8, unique=0;
    int sndable=1, rcvable=1, f=0, fs=8;
    int ftbreak=IEM_BNG_DEFAULTBREAKFLASHTIME, fthold=IEM_BNG_DEFAULTHOLDFLASHTIME;
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");


    if((argc == 14)&&IS_A_FLOAT(argv,0)
       &&IS_A_FLOAT(argv,1)&&IS_A_FLOAT(argv,2)
       &&IS_A_FLOAT(argv,3)
       &&IS_A_SYMBOL(argv,4)&&IS_A_SYMBOL(argv,5)
       &&IS_A_SYMBOL(argv,6)&&IS_A_FLOAT(argv,7)&&IS_A_FLOAT(argv,8)
       &&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10)&&IS_A_FLOAT(argv,11)
       &&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13))
    {
        a = (int)atom_getintarg(0, argc, argv);
	fthold = (int)atom_getintarg(1, argc, argv);
        ftbreak = (int)atom_getintarg(2, argc, argv);
	init = (int)atom_getintarg(3, argc, argv);
        snd = atom_getsymbolarg(4, argc, argv);
	rcv = atom_getsymbolarg(5, argc, argv);
	lab = atom_getsymbolarg(6, argc, argv);
	ldx = (int)atom_getintarg(7, argc, argv);
	ldy = (int)atom_getintarg(8, argc, argv);
	f = ((int)atom_getintarg(9, argc, argv))&255;
	unique = ((int)atom_getintarg(9, argc, argv))&1792;

        fs = (int)atom_getintarg(10, argc, argv);
        bcol = (int)atom_getintarg(11, argc, argv);
	fcol = (int)atom_getintarg(12, argc, argv);
        lcol = (int)atom_getintarg(13, argc, argv);
    }
    x->x_flashed = 0;
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(init != 0) init = 1;
    x->x_loadinit = init;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    reconvert2unique(snd, unique & 256);
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;

    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a);
    bng_check_minmax(x, ftbreak, fthold);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(fcol < 0)
    {
        fcol = -1 - fcol;
	x->x_fcol = ((fcol & 0x3f000) << 6)|((fcol & 0xfc0) << 4)|((fcol & 0x3f) << 2);
    }
    else
    {
	fcol = iem_gui_modulo_color(fcol);
	x->x_fcol = iem_gui_color_hex[fcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }

    x->x_clock_hld = clock_new(x, (t_method)bng_tick_hld);
    x->x_clock_brk = clock_new(x, (t_method)bng_tick_brk);
    x->x_selected = 0;
    x->x_shiftdown = 0;
    outlet_new(&x->x_obj, &s_bang);
    return (x);
}

static void bng_ff(t_bng *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
    clock_free(x->x_clock_brk);
    clock_free(x->x_clock_hld);
}

void bng_setup(void)
{
    bng_class = class_new(gensym("bng"), (t_newmethod)bng_new,
			  (t_method)bng_ff, sizeof(t_bng), 0, A_GIMME, 0);
    class_addbang(bng_class, bng_bang);
    class_addfloat(bng_class, bng_float);
    class_addsymbol(bng_class, bng_symbol);
    class_addpointer(bng_class, bng_pointer);
    class_addlist(bng_class, bng_list);
    class_addanything(bng_class, bng_anything);
    class_addmethod(bng_class, (t_method)bng_click, gensym("click"),
		    A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(bng_class, (t_method)bng_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_loadbang, gensym("loadbang"), 0);
    class_addmethod(bng_class, (t_method)bng_size, gensym("size"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_flashtime, gensym("flashtime"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_color, gensym("color"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(bng_class, (t_method)bng_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(bng_class, (t_method)bng_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(bng_class, (t_method)bng_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(bng_class, (t_method)bng_init, gensym("init"), A_FLOAT, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    bng_widgetbehavior.w_getrectfn = bng_getrect;
    bng_widgetbehavior.w_displacefn = bng_displace;
    bng_widgetbehavior.w_selectfn = bng_select;
    bng_widgetbehavior.w_activatefn = NULL;
    bng_widgetbehavior.w_deletefn = bng_delete;
    bng_widgetbehavior.w_visfn = bng_vis;
#if (PD_VERSION_MINOR > 31)
    bng_widgetbehavior.w_clickfn = bng_newclick;
    bng_widgetbehavior.w_propertiesfn = bng_properties;
#endif
    bng_widgetbehavior.w_savefn = bng_save;
    class_setwidget(bng_class, &bng_widgetbehavior);
    class_sethelpsymbol(bng_class, gensym("bng"));
}

/* ------------ horicontal  slider ----------------------- */

typedef struct _hslider
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_w;
     int      x_h;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     int      x_pos;
     int      x_val;
     int      x_center;
     int      x_thick;
     int      x_lcol;
     int      x_fcol;
     int      x_bcol;
     int      x_unique;
     int      x_lin0_log1;
     int      x_loadinit;
     int      x_steady;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_finemoved;
     int      x_selected;
     int      x_shiftdown;
     double   x_min;
     double   x_max;
     double   x_k;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
} t_hslider;

t_widgetbehavior hslider_widgetbehavior;
static t_class *hslider_class;

/* widget helper functions */

static void hslider_update(t_hslider *x, t_glist *glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if (glist_isvisible(glist))
    {
	int r = x->x_obj.te_xpos + (x->x_val + 50)/100;
	sys_vgui(".x%x.c coords %xKNOB %d %d %d %d\n",
		 canvas, x, r, x->x_obj.te_ypos+1,
		 r, x->x_obj.te_ypos + x->x_h);
	if(x->x_val == x->x_center)
	{
	    if(!x->x_thick)
	    {
		sys_vgui(".x%x.c itemconfigure %xKNOB -width 7\n", canvas, x);
                x->x_thick = 1;
	    }
	}
	else
	{
            if(x->x_thick)
	    {
		sys_vgui(".x%x.c itemconfigure %xKNOB -width 3\n", canvas, x);
                x->x_thick = 0;
	    }
	}
    }
}

static void hslider_draw(t_hslider *x, t_glist *glist, int firsttime)
{
    int r = x->x_obj.te_xpos + (x->x_val + 50)/100;
    t_canvas *canvas=glist_getcanvas(glist);

    if (firsttime)
    {
	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos-3, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_w+2, x->x_obj.te_ypos + x->x_h,
		 x->x_bcol, x);
	sys_vgui(".x%x.c create line %d %d %d %d -width 3 -fill #%6.6x -tags %xKNOB\n",
		 canvas, r, x->x_obj.te_ypos+1, r,
		 x->x_obj.te_ypos + x->x_h, x->x_fcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize, x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize, x->x_lcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas, x->x_obj.te_xpos-3, x->x_obj.te_ypos + x->x_h-1,
		 x->x_obj.te_xpos+4, x->x_obj.te_ypos + x->x_h, x, 0);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas, x->x_obj.te_xpos-3, x->x_obj.te_ypos,
		 x->x_obj.te_xpos+4, x->x_obj.te_ypos+1, x, 0);
    }
    else
    {
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x,
		 x->x_obj.te_xpos-3, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_w+2, x->x_obj.te_ypos + x->x_h);
        sys_vgui(".x%x.c coords %xKNOB %d %d %d %d\n",
		 canvas, x, r, x->x_obj.te_ypos+1,
		 r, x->x_obj.te_ypos + x->x_h);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
	sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos-3, x->x_obj.te_ypos + x->x_h-1,
		 x->x_obj.te_xpos+4, x->x_obj.te_ypos + x->x_h);
	sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos-3, x->x_obj.te_ypos,
		 x->x_obj.te_xpos+4, x->x_obj.te_ypos+1);
    }
}

static void hslider_erase(t_hslider* x,t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xKNOB\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
}

/* ------------------------ hsld widgetbehaviour----------------------------- */


static void hslider_getrect(t_gobj *z, t_glist *owner,
    int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_hslider* x = (t_hslider*)z;

    *xp1 = x->x_obj.te_xpos - 3;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_w + 2;
    *yp2 = x->x_obj.te_ypos + x->x_h;
}

static void hslider_displace(t_gobj *z, t_glist *glist,
    int dx, int dy)
{
    t_hslider* x = (t_hslider*)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    hslider_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text*)z);
}

static void hslider_select(t_gobj *z, t_glist *glist, int selected)
{
    t_hslider *x = (t_hslider *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void hslider_delete(t_gobj *z, t_glist *glist)
{
    t_hslider *x = (t_hslider *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void hslider_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_hslider *x = (t_hslider *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 hslider_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 hslider_erase(x, glist);
         rtext_free(y);
    }
}

static void hslider_save(t_gobj *z, t_binbuf *b)
{
    t_hslider *x = (t_hslider *)z;
    int bcol, fcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    fcol = -1 - (((0xfc0000 & x->x_fcol) >> 6)|((0xfc00 & x->x_fcol) >> 4)|((0xfc & x->x_fcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiffiisssiiiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("hslider"), x->x_w, x->x_h,
		(float)x->x_min, (float)x->x_max,
                x->x_lin0_log1, x->x_loadinit,
		x->x_snd, x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8+x->x_unique, x->x_fontsize,
		bcol, fcol, lcol,
	        x->x_val, x->x_steady);
    binbuf_addv(b, ";");
}

void hslider_check_width(t_hslider *x, int w)
{
    if(w < IEM_SL_MINSIZE)
	w = IEM_SL_MINSIZE;
    x->x_w = w;
    x->x_center = (x->x_w-1)*50;
    if(x->x_val > (x->x_w*100 - 100))
    {
	x->x_pos = x->x_w*100 - 100;
	x->x_val = x->x_pos;
    }
    if(x->x_lin0_log1)
	x->x_k = log(x->x_max/x->x_min)/(double)(x->x_w - 1);
    else
	x->x_k = (x->x_max - x->x_min)/(double)(x->x_w - 1);
}

void hslider_check_minmax(t_hslider *x, double min, double max)
{
    if(x->x_lin0_log1)
    {
	if((min == 0.0)&&(max == 0.0))
            max = 1.0;
	if(max > 0.0)
	{
	    if(min <= 0.0)
                min = 0.01*max;
	}
	else
	{
            if(min > 0.0)
                max = 0.01*min;
	}
    }
    x->x_min = min;
    x->x_max = max;
    if(x->x_lin0_log1)
        x->x_k = log(x->x_max/x->x_min)/(double)(x->x_w - 1);
    else
	x->x_k = (x->x_max - x->x_min)/(double)(x->x_w - 1);
}

static void hslider_properties(t_gobj *z, t_glist *owner)
{
    t_hslider *x = (t_hslider *)z;
    char buf[600];
    int steady=-23;

    if(x->x_steady)
	steady = -22;
    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);

    sprintf(buf, "pdtk_iemgui_dialog %%s HSLIDER \
	    --------dimensions(pix)(pix):-------- %g %g width: %g %g height: \
	    -----------output-range:----------- %g left: %g right: %g \
	    %g lin log %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_w, (float)IEM_SL_MINSIZE, (float)x->x_h, (float)IEM_GUI_MINSIZE,
	    x->x_min, x->x_max, 0.0,/*no_schedule*/
	    (float)x->x_lin0_log1, (float)x->x_loadinit, (float)steady,/*no multi, but iem-characteristic*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
	    0xffffff & x->x_bcol, 0xffffff & x->x_fcol, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void hslider_bang(t_hslider *x)
{
    double out;

    if(x->x_lin0_log1)
	out = x->x_min*exp(x->x_k*(double)(x->x_val)*0.01);
    else
	out = (double)(x->x_val)*0.01*x->x_k + x->x_min;
    if((out < 1.0e-10)&&(out > -1.0e-10))
        out = 0.0;

    outlet_float(x->x_obj.ob_outlet, out);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_float(x->x_snd->s_thing, out);
}

static void hslider_dialog(t_hslider *x, t_symbol *s, int argc, t_atom *argv)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int w = (int)atom_getintarg(0, argc, argv);
    int h = (int)atom_getintarg(1, argc, argv);
    double min = (double)atom_getfloatarg(2, argc, argv);
    double max = (double)atom_getfloatarg(3, argc, argv);
    int lilo = (int)atom_getintarg(4, argc, argv);
    int init = (int)atom_getintarg(5, argc, argv);
    int steady = (int)atom_getintarg(6, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int fcol = (int)atom_getintarg(15, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    /*post("w= %d, h= %d, min= %.3f, max= %.3f", w, h, min, max);
    post("lilo= %d, init= %d", lilo, init);
    post("snd= %s, rcv= %s", snd->s_name, rcv->s_name);
    post("lab= %s, ldx= %d, ldy= %d", lab->s_name, ldx, ldy);
    post("bcol= %s, fcol= %s, lcol= %s", iem_gui_color_str[bcol], iem_gui_color_str[fcol],
	 iem_gui_color_str[lcol]);*/

    if(lilo != 0) lilo = 1;
    x->x_lin0_log1 = lilo;
    if(init != 0) init = 1;
    x->x_loadinit = init;
    if(steady == 3)
	x->x_steady = 0;
    else
	x->x_steady = 1;
    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(snd))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;

    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }

    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_h = iem_gui_clip_size(h);
    hslider_check_width(x, w);
    hslider_check_minmax(x, min, max);
    x->x_lcol = lcol & 0xffffff;
    x->x_fcol = fcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, x->x_lab->s_name);
    sys_vgui(".x%x.c itemconfigure %xKNOB -fill #%6.6x\n", canvas, x, x->x_fcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x, x->x_bcol);
    hslider_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(canvas, (t_text*)x);
}

static void hslider_motion(t_hslider *x, t_floatarg dx, t_floatarg dy)
{
    int old = x->x_val;

    if(x->x_finemoved)
        x->x_pos += (int)dx;
    else
	x->x_pos += 100*(int)dx;
    x->x_val = x->x_pos;
    if(x->x_val > (100*x->x_w - 100))
    {
	x->x_val = 100*x->x_w - 100;
        x->x_pos += 50;
	/*x->x_pos /= 100;
	 x->x_pos *= 100;*/
        x->x_pos -= x->x_pos%100;
    }
    if(x->x_val < 0)
    {
	x->x_val = 0;
	x->x_pos -= 50;
	/*x->x_pos /= 100;
	 x->x_pos *= 100;*/
        x->x_pos -= x->x_pos%100;
    }
    if(old != x->x_val)
    {
	hslider_update(x, x->x_glist);
	hslider_bang(x);
    }
}

static void hslider_click(t_hslider *x, t_floatarg xpos, t_floatarg ypos,
		       t_floatarg shift, t_floatarg ctrl, t_floatarg alt)
{
    if(!x->x_steady)
	x->x_val = (int)(100.0 * (xpos - x->x_obj.te_xpos));
    if(x->x_val > (100*x->x_w - 100))
	x->x_val = 100*x->x_w - 100;
    if(x->x_val < 0)
	x->x_val = 0;
    x->x_pos = x->x_val;
    hslider_update(x, x->x_glist);
    hslider_bang(x);
#if (PD_VERSION_MINOR > 31)
    glist_grab(x->x_glist, &x->x_obj.te_g, (t_glistmotionfn)hslider_motion,
	       0, xpos, ypos);
#else
    glist_grab(x->x_glist, &x->x_obj.te_g, xpos, ypos);
#endif
}

#if (PD_VERSION_MINOR > 31)
static int hslider_newclick(t_gobj *z, struct _glist *glist,
    int xpix, int ypix, int shift, int alt, int dbl, int doit)
{
    t_hslider* x = (t_hslider *)z;

    if(doit)
    {
	hslider_click( x, (t_floatarg)xpix, (t_floatarg)ypix, (t_floatarg)shift,
		    0, (t_floatarg)alt);
	if(shift)
	    x->x_finemoved = 1;
	else
            x->x_finemoved = 0;
    }
    return (1);
}
#endif

static void hslider_set(t_hslider *x, t_floatarg f)
{
    double g;

    if(f > x->x_max)
	f = x->x_max;
    if(f < x->x_min)
	f = x->x_min;
    if(x->x_lin0_log1)
	g = log(f/x->x_min)/x->x_k;
    else
	g = (f - x->x_min) / x->x_k;
    x->x_val = (int)(100.0*g + 0.49999);
    x->x_pos = x->x_val;
    hslider_update(x, x->x_glist);
}

static void hslider_size(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    hslider_check_width(x, (int)atom_getintarg(0, ac, av));
    x->x_h = iem_gui_clip_size((int)atom_getintarg(1, ac, av));
    if(glist_isvisible(x->x_glist))
    {
        hslider_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void hslider_delta(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        hslider_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void hslider_pos(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        hslider_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void hslider_range(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    hslider_check_minmax(x, (double)atom_getfloatarg(0, ac, av),
			 (double)atom_getfloatarg(1, ac, av));
}

static void hslider_color(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i,j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_fcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_fcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(2, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    {
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas,
	     x, x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
	sys_vgui(".x%x.c itemconfigure %xKNOB -fill #%6.6x\n", canvas,
		 x, x->x_fcol);
	sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
		 x, x->x_bcol);
    }
}

static void hslider_send(t_hslider *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void hslider_receive(t_hslider *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void hslider_label(t_hslider *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    x->x_lab = s;
    if(glist_isvisible(x->x_glist))
    {
	if(!strcmp(s->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, s->s_name);
    }
}

static void hslider_label_pos(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c coords %xLABEL %d %d\n", glist_getcanvas(x->x_glist), x,
		 x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
    }
}

static void hslider_label_font(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
		 glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
    }
}

static void hslider_log(t_hslider *x)
{
    x->x_lin0_log1 = 1;
    hslider_check_minmax(x, x->x_min, x->x_max);
}

static void hslider_lin(t_hslider *x)
{
    x->x_lin0_log1 = 0;
    x->x_k = (x->x_max - x->x_min)/(double)(x->x_w - 1);
}

static void hslider_init(t_hslider *x, t_floatarg f)
{
    x->x_loadinit = (f==0.0)?0:1;
}

static void hslider_steady(t_hslider *x, t_floatarg f)
{
    x->x_steady = (f==0.0)?0:1;
}

static void hslider_float(t_hslider *x, t_floatarg f)
{
    hslider_set(x, f);
    hslider_bang(x);
}

static void hslider_loadbang(t_hslider *x)
{
    if(!sys_noloadbang && x->x_loadinit)
    {
	hslider_update(x, x->x_glist);
	hslider_bang(x);
    }
}

static void hslider_list(t_hslider *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto hsliderlistelse;
	else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;

		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
		    {
			hslider_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
	return;
    }
hsliderlistelse:
    if(IS_A_FLOAT(av,0))
	hslider_float(x, atom_getfloatarg(0, ac, av));
}

static void *hslider_new(t_symbol *s, int argc, t_atom *argv)
{
    t_hslider *x = (t_hslider *)pd_new(hslider_class);
    int bcol= -262144;
    int fcol= -1;
    int lcol= -1;
    int w=IEM_SL_DEFAULTSIZE, h=IEM_GUI_DEFAULTSIZE;
    int lilo=0, init=0, ldx=20, ldy=8, f=0, fs=8;
    int sndable=1, rcvable=1, v=0, steady=1, unique=0;
    double min=0.0, max=(double)(IEM_SL_DEFAULTSIZE-1);
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if(((argc == 17)||(argc == 18))&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)
       &&IS_A_FLOAT(argv,4)&&IS_A_FLOAT(argv,5)
       &&IS_A_SYMBOL(argv,6)&&IS_A_SYMBOL(argv,7)
       &&IS_A_SYMBOL(argv,8)&&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10)
       &&IS_A_FLOAT(argv,11)&&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13)
       &&IS_A_FLOAT(argv,14)&&IS_A_FLOAT(argv,15)&&IS_A_FLOAT(argv,16))
    {
        w = (int)atom_getintarg(0, argc, argv);
	h = (int)atom_getintarg(1, argc, argv);
	min = (double)atom_getfloatarg(2, argc, argv);
	max = (double)atom_getfloatarg(3, argc, argv);
	lilo = (int)atom_getintarg(4, argc, argv);
	init = (int)atom_getintarg(5, argc, argv);
        snd = atom_getsymbolarg(6, argc, argv);
	rcv = atom_getsymbolarg(7, argc, argv);
	lab = atom_getsymbolarg(8, argc, argv);
	ldx = (int)atom_getintarg(9, argc, argv);
	ldy = (int)atom_getintarg(10, argc, argv);
        f = ((int)atom_getintarg(11, argc, argv))&255;
	unique = ((int)atom_getintarg(11, argc, argv))&1792;
        fs = (int)atom_getintarg(12, argc, argv);
        bcol = (int)atom_getintarg(13, argc, argv);
	fcol = (int)atom_getintarg(14, argc, argv);
        lcol = (int)atom_getintarg(15, argc, argv);
	v = (int)atom_getintarg(16, argc, argv);
    }
    if((argc == 18)&&IS_A_FLOAT(argv,17))
	steady = (int)atom_getintarg(17, argc, argv);
    if(init != 0) init = 1;
    x->x_loadinit = init;
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(x->x_loadinit)
	x->x_val = v;
    else
        x->x_val = 0;
    x->x_pos = x->x_val;
    if(lilo != 0) lilo = 1;
    x->x_lin0_log1 = lilo;
    if(steady != 0) steady = 1;
    x->x_steady = steady;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    reconvert2unique(snd, unique & 256);
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_h = iem_gui_clip_size(h);
    hslider_check_width(x, w);
    hslider_check_minmax(x, min, max);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(fcol < 0)
    {
        fcol = -1 - fcol;
	x->x_fcol = ((fcol & 0x3f000) << 6)|((fcol & 0xfc0) << 4)|((fcol & 0x3f) << 2);
    }
    else
    {
	fcol = iem_gui_modulo_color(fcol);
	x->x_fcol = iem_gui_color_hex[fcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_finemoved = 0;
    x->x_selected = 0;
    x->x_shiftdown = 0;
    x->x_thick = 0;
    outlet_new(&x->x_obj, &s_float);
    return (x);
}

static void hslider_free(t_hslider *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void hslider_setup(void)
{
    hslider_class = class_new(gensym("hsl"), (t_newmethod)hslider_new,
			      (t_method)hslider_free, sizeof(t_hslider), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)hslider_new, gensym("hslider"), A_GIMME, 0);
    class_addbang(hslider_class,hslider_bang);
    class_addfloat(hslider_class,hslider_float);
    class_addlist(hslider_class, hslider_list);
    class_addmethod(hslider_class, (t_method)hslider_click, gensym("click"),
		    A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(hslider_class, (t_method)hslider_motion, gensym("motion"),
		    A_FLOAT, A_FLOAT, 0);
    class_addmethod(hslider_class, (t_method)hslider_dialog, gensym("dialog"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_loadbang, gensym("loadbang"), 0);
    class_addmethod(hslider_class, (t_method)hslider_set, gensym("set"), A_FLOAT, 0);
    class_addmethod(hslider_class, (t_method)hslider_size, gensym("size"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_range, gensym("range"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_color, gensym("color"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(hslider_class, (t_method)hslider_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(hslider_class, (t_method)hslider_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(hslider_class, (t_method)hslider_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(hslider_class, (t_method)hslider_log, gensym("log"), 0);
    class_addmethod(hslider_class, (t_method)hslider_lin, gensym("lin"), 0);
    class_addmethod(hslider_class, (t_method)hslider_init, gensym("init"), A_FLOAT, 0);
    class_addmethod(hslider_class, (t_method)hslider_steady, gensym("steady"), A_FLOAT, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    hslider_widgetbehavior.w_getrectfn =    hslider_getrect;
    hslider_widgetbehavior.w_displacefn =   hslider_displace;
    hslider_widgetbehavior.w_selectfn =     hslider_select;
    hslider_widgetbehavior.w_activatefn =   NULL;
    hslider_widgetbehavior.w_deletefn =     hslider_delete;
    hslider_widgetbehavior.w_visfn =        hslider_vis;
#if (PD_VERSION_MINOR > 31)
    hslider_widgetbehavior.w_clickfn =      hslider_newclick;
    hslider_widgetbehavior.w_propertiesfn = hslider_properties;
#endif
    hslider_widgetbehavior.w_savefn =       hslider_save;
    class_setwidget(hslider_class, &hslider_widgetbehavior);
    class_sethelpsymbol(hslider_class, gensym("hslider"));
}

/* ------------- radiobut     gui-radiobutton ---------------------- */

t_widgetbehavior radiobut_widgetbehavior;
static t_class *radiobut_class;

typedef struct _radiobut
{
    t_object x_obj;
    t_glist *x_glist;
    int      x_size;
    int      x_ldx;
    int      x_ldy;
    char     x_font[16];
    int      x_fontsize;
    int      x_on;
    int      x_on_old;
    int      x_change;
    int      x_number;
    int      x_fcol;
    int      x_bcol;
    int      x_lcol;
    int      x_unique;
    int      x_loadinit;
    int      x_snd_able;
    int      x_rcv_able;
    int      x_selected;
    int      x_shiftdown;
    t_symbol *x_snd;
    t_symbol *x_rcv;
    t_symbol *x_lab;
    t_atom   x_at[2];
} t_radiobut;

/* widget helper functions */

void radiobut_update(t_radiobut *x, t_glist *glist)
{
    if(glist_isvisible(glist))
    {
	t_canvas *canvas=glist_getcanvas(glist);

	sys_vgui(".x%x.c itemconfigure %xBUT%d -fill #%6.6x -outline #%6.6x\n", canvas, x, x->x_on_old,
		 x->x_bcol, x->x_bcol);
	sys_vgui(".x%x.c itemconfigure %xBUT%d -fill #%6.6x -outline #%6.6x\n", canvas, x, x->x_on,
		 x->x_fcol, x->x_fcol);
    }
}

void radiobut_draw(t_radiobut *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);
    int n=x->x_number, i, dx=x->x_size, s4=dx/4;
    int yy11=x->x_obj.te_ypos, yy12=yy11+dx;
    int yy21=yy11+s4, yy22=yy12-s4;
    int xx11b=x->x_obj.te_xpos, xx11=x->x_obj.te_xpos, xx21=xx11b+s4;
    int xx22=xx11b+dx-s4;

    if(firsttime)
    {
	for(i=0; i<n; i++)
	{
	    sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE%d\n",
		     canvas, xx11, yy11, xx11+dx, yy12,
		     x->x_bcol, x, i);
	    sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -outline #%6.6x -tags %xBUT%d\n",
		     canvas, xx21, yy21, xx22, yy22,
		     (x->x_on==i)?x->x_fcol:x->x_bcol,
		     (x->x_on==i)?x->x_fcol:x->x_bcol, x, i);
	    xx11 += dx;
	    xx21 += dx;
	    xx22 += dx;
	}
	if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, xx11b+x->x_ldx, yy11+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
	else
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, xx11b+x->x_ldx, yy11+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas, xx11b, yy12-1, xx11b + IOWIDTH, yy12, x, 0);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas, xx11b, yy11, xx11b + IOWIDTH, yy11+1, x, 0);
    }
    else
    {
	xx11 = xx11b;
	xx21=xx11b+s4;
	xx22=xx11b+dx-s4;
	for(i=0; i<n; i++)
	{
	    sys_vgui(".x%x.c coords %xBASE%d %d %d %d %d\n",
		     canvas, x, i, xx11, yy11, xx11+dx, yy12);
	    sys_vgui(".x%x.c coords %xBUT%d %d %d %d %d\n",
		     canvas, x, i, xx21, yy21, xx22, yy22);
	    xx11 += dx;
	    xx21 += dx;
	    xx22 += dx;
	}
	sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, xx11b+x->x_ldx, yy11+x->x_ldy);
	sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0, xx11b, yy12-1, xx11b + IOWIDTH, yy12);
	sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0, xx11b, yy11, xx11b + IOWIDTH, yy11+1);
    }
}

void radiobut_erase(t_radiobut* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);
    int n=x->x_number, i;

    for(i=0; i<n; i++)
    {
	sys_vgui(".x%x.c delete %xBASE%d\n", canvas, x, i);
	sys_vgui(".x%x.c delete %xBUT%d\n", canvas, x, i);
    }
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void radiobut_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_radiobut *x = (t_radiobut *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size*x->x_number;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void radiobut_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_radiobut *x = (t_radiobut *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    radiobut_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void radiobut_select(t_gobj *z, t_glist *glist, int selected)
{
    t_radiobut *x = (t_radiobut *)z;
    t_canvas *canvas=glist_getcanvas(glist);
    int n=x->x_number, i;

    x->x_selected = selected;
    if(selected)
    {
	pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	for(i=0; i<n; i++)
	{
	    sys_vgui(".x%x.c itemconfigure %xBASE%d -outline #%6.6x\n", canvas, x, i,
		     IEM_GUI_COLOR_SELECTED);
	}
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
	pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	for(i=0; i<n; i++)
	{
	    sys_vgui(".x%x.c itemconfigure %xBASE%d -outline #%6.6x\n", canvas, x, i,
		     IEM_GUI_COLOR_NORMAL);
	}
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x,
		 x->x_lcol);
    }
}

static void radiobut_delete(t_gobj *z, t_glist *glist)
{
    t_radiobut *x = (t_radiobut *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void radiobut_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_radiobut *x = (t_radiobut *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 radiobut_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 radiobut_erase(x, glist);
         rtext_free(y);
    }
}

static void radiobut_save(t_gobj *z, t_binbuf *b)
{
    t_radiobut *x = (t_radiobut *)z;
    int bcol, fcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    fcol = -1 - (((0xfc0000 & x->x_fcol) >> 6)|((0xfc00 & x->x_fcol) >> 4)|((0xfc & x->x_fcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiiisssiiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("radiobut"), x->x_size,
		x->x_change, x->x_loadinit, x->x_number,
		x->x_snd, x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
		(int)(x->x_font[0]-'a'+1)/8+x->x_unique, x->x_fontsize,
		bcol, fcol, lcol, x->x_on);
    binbuf_addv(b, ";");
}

static void radiobut_properties(t_gobj *z, t_glist *owner)
{
    t_radiobut *x = (t_radiobut *)z;
    char buf[600];

    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s RADIO-BUTTON \
	    ----------dimensions(pix):----------- %g %g size: 0 0 empty \
	    empty 0.0 empty 0.0 empty %g \
	    %g new-only new&old %g %g \
	    %s %s \
	    %s %g %g \
	    %g %g \
	    %d %d %d\n",
	    (float)x->x_size, (float)IEM_GUI_MINSIZE,
	    0.0,/*no_schedule*/
	    (float)x->x_change, (float)x->x_loadinit, (float)x->x_number,
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
	    (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
	    0xffffff & x->x_bcol, 0xffffff & x->x_fcol, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void radiobut_dialog(t_radiobut *x, t_symbol *s, int argc, t_atom *argv)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int a = (int)atom_getintarg(0, argc, argv);
    int chg = (int)atom_getintarg(4, argc, argv);
    int init = (int)atom_getintarg(5, argc, argv);
    int num = (int)atom_getintarg(6, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int fcol = (int)atom_getintarg(15, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1, i;

    if(init != 0) init = 1;
    x->x_loadinit = init;
    if(chg != 0) chg = 1;
    x->x_change = chg;
    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(snd))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
	    x->x_rcv = rcv;
	    pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }
    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_size = iem_gui_clip_size(a);
    x->x_lcol = lcol & 0xffffff;
    x->x_fcol = fcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    if(x->x_number != num)
    {
	radiobut_erase(x, x->x_glist);
	x->x_number = num;
	if(x->x_on >= x->x_number)
	{
	    x->x_on = x->x_number - 1;
            x->x_on_old = x->x_on;
	}
	radiobut_draw(x, x->x_glist, 1);
    }
    else
    {
	if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		     x->x_lab->s_name);
	for(i=0; i<num; i++)
	{
	    sys_vgui(".x%x.c itemconfigure %xBASE%d -fill #%6.6x\n", canvas, x, i,
		     x->x_bcol);
	    sys_vgui(".x%x.c itemconfigure %xBUT%d -fill #%6.6x -outline #%6.6x\n", canvas, x, i,
		     (x->x_on==i)?x->x_fcol:x->x_bcol,
		     (x->x_on==i)?x->x_fcol:x->x_bcol);
	}
	radiobut_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(canvas, (t_text*)x);
    }
}

static void radiobut_set(t_radiobut *x, t_floatarg f)
{
    int i=(int)f;
    int old;

    if(i < 0)
	i = 0;
    if(i >= x->x_number)
	i = x->x_number-1;
    if(x->x_on != x->x_on_old)
    {
	old = x->x_on_old;
	x->x_on_old = x->x_on;
	x->x_on = i;
	radiobut_update(x, x->x_glist);
	x->x_on_old = old;
    }
    else
    {
	x->x_on = i;
	radiobut_update(x, x->x_glist);
    }
}

static void radiobut_float(t_radiobut *x, t_floatarg f)
{
    int i=(int)f;

    if(i < 0)
	i = 0;
    if(i >= x->x_number)
	i = x->x_number-1;

    if((x->x_change)&&(i != x->x_on_old))
    {
	SETFLOAT(x->x_at, (float)x->x_on_old);
	SETFLOAT(x->x_at+1, 0.0);
	outlet_list(x->x_obj.ob_outlet, &s_list, 2, x->x_at);
	if(x->x_snd_able && x->x_snd->s_thing)
	    pd_list(x->x_snd->s_thing, &s_list, 2, x->x_at);
    }
    if(x->x_on != x->x_on_old)
	x->x_on_old = x->x_on;
    x->x_on = i;
    radiobut_update(x, x->x_glist);
    x->x_on_old = x->x_on;
    SETFLOAT(x->x_at, (float)x->x_on);
    SETFLOAT(x->x_at+1, 1.0);
    outlet_list(x->x_obj.ob_outlet, &s_list, 2, x->x_at);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_list(x->x_snd->s_thing, &s_list, 2, x->x_at);
}

static void radiobut_bang(t_radiobut *x)
{
    if((x->x_change)&&(x->x_on != x->x_on_old))
    {
        SETFLOAT(x->x_at, (float)x->x_on_old);
	SETFLOAT(x->x_at+1, 0.0);
	outlet_list(x->x_obj.ob_outlet, &s_list, 2, x->x_at);
	if(x->x_snd_able && x->x_snd->s_thing)
	    pd_list(x->x_snd->s_thing, &s_list, 2, x->x_at);
    }
    x->x_on_old = x->x_on;
    SETFLOAT(x->x_at, (float)x->x_on);
    SETFLOAT(x->x_at+1, 1.0);
    outlet_list(x->x_obj.ob_outlet, &s_list, 2, x->x_at);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_list(x->x_snd->s_thing, &s_list, 2, x->x_at);
}

static void radiobut_click(t_radiobut *x, t_floatarg xpos, t_floatarg ypos, t_floatarg shift, t_floatarg ctrl, t_floatarg alt)
{
    int xx = (int)xpos - (int)x->x_obj.te_xpos;

    radiobut_float(x, (float)(xx / x->x_size));
}

#if (PD_VERSION_MINOR > 31)
static int radiobut_newclick(t_gobj *z, struct _glist *glist, int xpix, int ypix, int shift, int alt, int dbl, int doit)
{
    if(doit)
	radiobut_click((t_radiobut *)z, (t_floatarg)xpix, (t_floatarg)ypix, (t_floatarg)shift, 0, (t_floatarg)alt);
    return (1);
}
#endif

static void radiobut_loadbang(t_radiobut *x)
{
    if(!sys_noloadbang && x->x_loadinit)
	radiobut_bang(x);
}

static void radiobut_number(t_radiobut *x, t_floatarg num)
{
    int n=(int)num;

    if(n < 1)
	n = 1;
    if(n > 128)
	n = 128;
    if(n != x->x_number)
    {
	radiobut_erase(x, x->x_glist);
	x->x_number = n;
	if(x->x_on >= x->x_number)
	    x->x_on = x->x_number - 1;
	x->x_on_old = x->x_on;
        if(glist_isvisible(x->x_glist))
	    radiobut_draw(x, x->x_glist, 1);
    }
}

static void radiobut_size(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_size = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    if(glist_isvisible(x->x_glist))
    {
        radiobut_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void radiobut_delta(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        radiobut_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void radiobut_pos(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        radiobut_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void radiobut_color(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int n=x->x_number, i, j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_fcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_fcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(2, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
	for(i=0; i<n; i++)
	{
	    sys_vgui(".x%x.c itemconfigure %xBASE%d -fill #%6.6x\n",
		     canvas, x, i, x->x_bcol);
	    sys_vgui(".x%x.c itemconfigure %xBUT%d -fill #%6.6x -outline #%6.6x\n", canvas, x, i,
		     (x->x_on==i)?x->x_fcol:x->x_bcol,
		     (x->x_on==i)?x->x_fcol:x->x_bcol);
	}
    }
}

static void radiobut_send(t_radiobut *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void radiobut_receive(t_radiobut *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
	    x->x_rcv = s;
	    pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void radiobut_label(t_radiobut *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    x->x_lab = s;
    if(glist_isvisible(x->x_glist))
    {
	if(!strcmp(s->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, s->s_name);
    }
}

static void radiobut_label_pos(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 glist_getcanvas(x->x_glist), x,
		 x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
    }
}

static void radiobut_label_font(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
    {
    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
    }
}

static void radiobut_init(t_radiobut *x, t_floatarg f)
{
    x->x_loadinit = (f==0.0)?0:1;
}

static void radiobut_double_change(t_radiobut *x)
{
    x->x_change = 1;
}

static void radiobut_single_change(t_radiobut *x)
{
    x->x_change = 0;
}

static void radiobut_list(t_radiobut *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto radiobutlistelse;
	else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;

		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
		    {
			radiobut_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
	return;
    }
radiobutlistelse:
    if(IS_A_FLOAT(av,0))
	radiobut_float(x, atom_getfloatarg(0, ac, av));
}

static void *radiobut_new(t_symbol *s, int argc, t_atom *argv)
{
    t_radiobut *x = (t_radiobut *)pd_new(radiobut_class);
    int bcol= -262144;
    int fcol= -1;
    int lcol= -1;
    int a=IEM_GUI_DEFAULTSIZE, unique=0;
    int init=0, ldx=8, ldy=-8, on=0, f=0, fs=8;
    int sndable=1, rcvable=1, chg=1, num=8;
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if((argc == 15)&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)&&IS_A_FLOAT(argv,2)
       &&IS_A_FLOAT(argv,3)&&IS_A_SYMBOL(argv,4)&&IS_A_SYMBOL(argv,5)
       &&IS_A_SYMBOL(argv,6)&&IS_A_FLOAT(argv,7)&&IS_A_FLOAT(argv,8)
       &&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10)&&IS_A_FLOAT(argv,11)
       &&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13)&&IS_A_FLOAT(argv,14))
    {
	a = (int)atom_getintarg(0, argc, argv);
	chg = (int)atom_getintarg(1, argc, argv);
	init = (int)atom_getintarg(2, argc, argv);
	num = (int)atom_getintarg(3, argc, argv);
	snd = atom_getsymbolarg(4, argc, argv);
	rcv = atom_getsymbolarg(5, argc, argv);
	lab = atom_getsymbolarg(6, argc, argv);
	ldx = (int)atom_getintarg(7, argc, argv);
	ldy = (int)atom_getintarg(8, argc, argv);
        f = ((int)atom_getintarg(9, argc, argv))&255;
	unique = ((int)atom_getintarg(9, argc, argv))&1792;
	fs = (int)atom_getintarg(10, argc, argv);
	bcol = (int)atom_getintarg(11, argc, argv);
	fcol = (int)atom_getintarg(12, argc, argv);
	lcol = (int)atom_getintarg(13, argc, argv);
	on = (int)atom_getintarg(14, argc, argv);
    }
    if(init != 0) init = 1;
    x->x_loadinit = init;
    if(num < 1)
	num = 1;
    if(num > 128)
	num = 128;
    x->x_number = num;
    if(on < 0)
	on = 0;
    if(on >= x->x_number)
	on = x->x_number - 1;
    if(x->x_loadinit)
	x->x_on = on;
    else
        x->x_on = 0;
    x->x_on_old = x->x_on;
    x->x_change = (chg==0)?0:1;
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    reconvert2unique(snd, unique & 256);
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(fcol < 0)
    {
        fcol = -1 - fcol;
	x->x_fcol = ((fcol & 0x3f000) << 6)|((fcol & 0xfc0) << 4)|((fcol & 0x3f) << 2);
    }
    else
    {
	fcol = iem_gui_modulo_color(fcol);
	x->x_fcol = iem_gui_color_hex[fcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    x->x_shiftdown = 0;
    outlet_new(&x->x_obj, &s_list);
    return (x);
}

static void radiobut_ff(t_radiobut *x)
{
    if(x->x_selected)
	pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void radiobut_setup(void)
{
    radiobut_class = class_new(gensym("rdb"), (t_newmethod)radiobut_new,
			       (t_method)radiobut_ff, sizeof(t_radiobut), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)radiobut_new, gensym("radiobut"), A_GIMME, 0);
    class_addcreator((t_newmethod)radiobut_new, gensym("radiobutton"), A_GIMME, 0);
    class_addbang(radiobut_class, radiobut_bang);
    class_addfloat(radiobut_class, radiobut_float);
    class_addlist(radiobut_class, radiobut_list);
    class_addmethod(radiobut_class, (t_method)radiobut_click, gensym("click"),
		    A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_loadbang, gensym("loadbang"), 0);
    class_addmethod(radiobut_class, (t_method)radiobut_set, gensym("set"), A_FLOAT, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_size, gensym("size"), A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_color, gensym("color"), A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_init, gensym("init"), A_FLOAT, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_number, gensym("number"), A_FLOAT, 0);
    class_addmethod(radiobut_class, (t_method)radiobut_single_change, gensym("single_change"), 0);
    class_addmethod(radiobut_class, (t_method)radiobut_double_change, gensym("double_change"), 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    radiobut_widgetbehavior.w_getrectfn = radiobut_getrect;
    radiobut_widgetbehavior.w_displacefn = radiobut_displace;
    radiobut_widgetbehavior.w_selectfn = radiobut_select;
    radiobut_widgetbehavior.w_activatefn = NULL;
    radiobut_widgetbehavior.w_deletefn = radiobut_delete;
    radiobut_widgetbehavior.w_visfn = radiobut_vis;
#if (PD_VERSION_MINOR > 31)
    radiobut_widgetbehavior.w_clickfn = radiobut_newclick;
    radiobut_widgetbehavior.w_propertiesfn = radiobut_properties;
#endif
    radiobut_widgetbehavior.w_savefn = radiobut_save;
    class_setwidget(radiobut_class, &radiobut_widgetbehavior);
    class_sethelpsymbol(radiobut_class, gensym("radiobutton"));
}

/* --------------- toggle     gui-toggle ------------------------- */

t_widgetbehavior toggle_widgetbehavior;
static t_class *toggle_class;

typedef struct _toggle
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_size;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     float    x_on;
     float    x_nonzero;
     int      x_fcol;
     int      x_bcol;
     int      x_lcol;
     int      x_unique;
     int      x_loadinit;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_selected;
     int      x_shiftdown;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
} t_toggle;

/* widget helper functions */

void toggle_update(t_toggle *x, t_glist *glist)
{
    if(glist_isvisible(glist))
    {
        t_canvas *canvas=glist_getcanvas(glist);

        sys_vgui(".x%x.c itemconfigure %xX1 -fill #%6.6x\n", canvas, x,
		 (x->x_on!=0.0)?x->x_fcol:x->x_bcol);
	sys_vgui(".x%x.c itemconfigure %xX2 -fill #%6.6x\n", canvas, x,
		 (x->x_on!=0.0)?x->x_fcol:x->x_bcol);
    }
}

void toggle_draw(t_toggle *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);
    int w=1, xx=x->x_obj.te_xpos, yy=x->x_obj.te_ypos;

    if(x->x_size >= 30)
	w = 2;
    if(x->x_size >= 60)
	w = 3;
    if(firsttime)
    {
        sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, xx, yy, xx + x->x_size, yy + x->x_size,
		 x->x_bcol, x);
	sys_vgui(".x%x.c create line %d %d %d %d -width %d -fill #%6.6x -tags %xX1\n",
		 canvas, xx+w, yy+w, xx + x->x_size-w, yy + x->x_size-w, w,
		 (x->x_on!=0.0)?x->x_fcol:x->x_bcol, x);
	sys_vgui(".x%x.c create line %d %d %d %d -width %d -fill #%6.6x -tags %xX2\n",
		 canvas, xx+w, yy + x->x_size-w, xx + x->x_size-w, yy+w, w,
		 (x->x_on!=0.0)?x->x_fcol:x->x_bcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas, xx, yy + x->x_size-1, xx + IOWIDTH, yy + x->x_size, x, 0);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas, xx, yy, xx + IOWIDTH, yy+1, x, 0);
    }
    else
    {
        sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, xx, yy, xx + x->x_size, yy + x->x_size);
        sys_vgui(".x%x.c itemconfigure %xX1 -width %d\n", canvas, x, w);
	sys_vgui(".x%x.c coords %xX1 %d %d %d %d\n",
		 canvas, x, xx+w, yy+w, xx + x->x_size-w, yy + x->x_size-w);
        sys_vgui(".x%x.c itemconfigure %xX2 -width %d\n", canvas, x, w);
	sys_vgui(".x%x.c coords %xX2 %d %d %d %d\n",
		 canvas, x, xx+w, yy + x->x_size-w, xx + x->x_size-w, yy+w);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, xx+x->x_ldx, yy+x->x_ldy);
        sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0, xx, yy + x->x_size-1, xx + IOWIDTH, yy + x->x_size);
        sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0, xx, yy, xx + IOWIDTH, yy+1);
    }
}

void toggle_erase(t_toggle* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xX1\n", canvas, x);
    sys_vgui(".x%x.c delete %xX2\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void toggle_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_toggle *x = (t_toggle *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void toggle_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_toggle *x = (t_toggle *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    toggle_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void toggle_select(t_gobj *z, t_glist *glist, int selected)
{
    t_toggle *x = (t_toggle *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void toggle_delete(t_gobj *z, t_glist *glist)
{
    t_toggle *x = (t_toggle *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void toggle_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_toggle *x = (t_toggle *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 toggle_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 toggle_erase(x, glist);
         rtext_free(y);
    }
}

static void toggle_save(t_gobj *z, t_binbuf *b)
{
    t_toggle *x = (t_toggle *)z;
    int bcol, fcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    fcol = -1 - (((0xfc0000 & x->x_fcol) >> 6)|((0xfc00 & x->x_fcol) >> 4)|((0xfc & x->x_fcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiisssiiiiiiiff", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("toggle"), x->x_size,
		x->x_loadinit,
                x->x_snd, x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8+x->x_unique, x->x_fontsize,
		bcol, fcol, lcol, x->x_on, x->x_nonzero);
    binbuf_addv(b, ";");
}

static void toggle_properties(t_gobj *z, t_glist *owner)
{
    t_toggle *x = (t_toggle *)z;
    char buf[600];
    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s TOGGLE \
	    ----------dimensions(pix):----------- %g %g size: 0 0 empty \
            -----------non-zero-value:----------- %g value: 0.0 empty %g \
	    -1 lin log %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_size, (float)IEM_GUI_MINSIZE,
	    x->x_nonzero, 1.0,/*non_zero-schedule*/
	    (float)x->x_loadinit, -1.0,/*no multi*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, 0xffffff & x->x_fcol, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void toggle_bang(t_toggle *x)
{
    x->x_on = (x->x_on==0.0)?x->x_nonzero:0.0;
    toggle_update(x, x->x_glist);
    outlet_float(x->x_obj.ob_outlet, x->x_on);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_float(x->x_snd->s_thing, x->x_on);
}

static void toggle_dialog(t_toggle *x, t_symbol *s, int argc, t_atom *argv)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int a = (int)atom_getintarg(0, argc, argv);
    float nonzero = (float)atom_getfloatarg(2, argc, argv);
    int init = (int)atom_getintarg(5, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int fcol = (int)atom_getintarg(15, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    if(nonzero == 0.0)
	nonzero = 1.0;
    x->x_nonzero = nonzero;
    if(x->x_on != 0.0)
        x->x_on = x->x_nonzero;
    if(init != 0) init = 1;
    x->x_loadinit = init;
    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(snd))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }
    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_size = iem_gui_clip_size(a);
    x->x_lcol = lcol & 0xffffff;
    x->x_fcol = fcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);
    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x,
	         x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xX1 -fill #%6.6x\n", canvas, x,
	     x->x_on?x->x_fcol:x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xX2 -fill #%6.6x\n", canvas, x,
	         x->x_on?x->x_fcol:x->x_bcol);
    toggle_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(canvas, (t_text*)x);
}

static void toggle_click(t_toggle *x, t_floatarg xpos, t_floatarg ypos, t_floatarg shift, t_floatarg ctrl, t_floatarg alt)
{
    toggle_bang(x);
}

#if (PD_VERSION_MINOR > 31)
static int toggle_newclick(t_gobj *z, struct _glist *glist, int xpix, int ypix, int shift, int alt, int dbl, int doit)
{
    if(doit)
	toggle_click((t_toggle *)z, (t_floatarg)xpix, (t_floatarg)ypix, (t_floatarg)shift, 0, (t_floatarg)alt);
    return (1);
}
#endif

static void toggle_set(t_toggle *x, t_floatarg f)
{
    x->x_on = f;
    if(f != 0.0)
        x->x_nonzero = f;
    toggle_update(x, x->x_glist);
}

static void toggle_float(t_toggle *x, t_floatarg f)
{
    toggle_set(x, f);
    outlet_float(x->x_obj.ob_outlet, x->x_on);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_float(x->x_snd->s_thing, x->x_on);
}

static void toggle_loadbang(t_toggle *x)
{
    if(!sys_noloadbang && x->x_loadinit)
        toggle_float(x, (float)x->x_on);
}

static void toggle_size(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_size = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    if(glist_isvisible(x->x_glist))
    {
        toggle_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void toggle_delta(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        toggle_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void toggle_pos(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        toggle_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void toggle_color(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i,j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_fcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_fcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(2, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    {
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas,
	     x, x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
        sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
	     x, x->x_bcol);
        sys_vgui(".x%x.c itemconfigure %xX1 -fill #%6.6x\n", canvas,
	     x, x->x_on?x->x_fcol:x->x_bcol);
        sys_vgui(".x%x.c itemconfigure %xX2 -fill #%6.6x\n", canvas,
	     x, x->x_on?x->x_fcol:x->x_bcol);
    }
}

static void toggle_send(t_toggle *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void toggle_receive(t_toggle *x, t_symbol *s)
{
    int rcvable=1;
    
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void toggle_label(t_toggle *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    x->x_lab = s;
    if(glist_isvisible(x->x_glist))
    {
        if(!strcmp(s->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, "");
        else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, s->s_name);
    }
}

static void toggle_label_font(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
        sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
}

static void toggle_label_pos(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
	sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 glist_getcanvas(x->x_glist), x,
	     x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
}

static void toggle_init(t_toggle *x, t_floatarg f)
{
    x->x_loadinit = (f==0.0)?0:1;
}

static void toggle_nonzero(t_toggle *x, t_floatarg f)
{
    if(f != 0.0)
        x->x_nonzero = f;
}

static void toggle_list(t_toggle *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto togglelistelse;
	else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;

		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
		    {
			toggle_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
	return;
    }
togglelistelse:
    if(IS_A_FLOAT(av,0))
	toggle_float(x, atom_getfloatarg(0, ac, av));
}

static void *toggle_new(t_symbol *s, int argc, t_atom *argv)
{
    t_toggle *x = (t_toggle *)pd_new(toggle_class);
    int bcol= -262144;
    int fcol= -1;
    int lcol= -1;
    int a=IEM_GUI_DEFAULTSIZE;
    int init=0, ldx=20, ldy=8, f=0, fs=8;
    int sndable=1, rcvable=1, unique=0;
    float on=0.0, nonzero=1.0;
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if(((argc == 13)||(argc == 14))&&IS_A_FLOAT(argv,0)
       &&IS_A_FLOAT(argv,1)
       &&IS_A_SYMBOL(argv,2)&&IS_A_SYMBOL(argv,3)
       &&IS_A_SYMBOL(argv,4)&&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)
       &&IS_A_FLOAT(argv,7)&&IS_A_FLOAT(argv,8)&&IS_A_FLOAT(argv,9)
       &&IS_A_FLOAT(argv,10)&&IS_A_FLOAT(argv,11)&&IS_A_FLOAT(argv,12))
    {
        a = (int)atom_getintarg(0, argc, argv);
	init = (int)atom_getintarg(1, argc, argv);
        snd = atom_getsymbolarg(2, argc, argv);
	rcv = atom_getsymbolarg(3, argc, argv);
	lab = atom_getsymbolarg(4, argc, argv);
	ldx = (int)atom_getintarg(5, argc, argv);
	ldy = (int)atom_getintarg(6, argc, argv);
        f = ((int)atom_getintarg(7, argc, argv))&255;
	unique = ((int)atom_getintarg(7, argc, argv))&1792;
        fs = (int)atom_getintarg(8, argc, argv);
        bcol = (int)atom_getintarg(9, argc, argv);
	fcol = (int)atom_getintarg(10, argc, argv);
	lcol = (int)atom_getintarg(11, argc, argv);
        on = (float)atom_getfloatarg(12, argc, argv);
    }
    if((argc == 14)&&IS_A_FLOAT(argv,13))
	nonzero = (float)atom_getfloatarg(13, argc, argv);
    if(init != 0) init = 1;
    x->x_loadinit = init;
    x->x_nonzero = (nonzero!=0.0)?nonzero:1.0;
    if(x->x_loadinit)
	x->x_on = (on!=0.0)?nonzero:0.0;
    else
        x->x_on = 0.0;
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    reconvert2unique(snd, unique & 256);
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(fcol < 0)
    {
        fcol = -1 - fcol;
	x->x_fcol = ((fcol & 0x3f000) << 6)|((fcol & 0xfc0) << 4)|((fcol & 0x3f) << 2);
    }
    else
    {
	fcol = iem_gui_modulo_color(fcol);
	x->x_fcol = iem_gui_color_hex[fcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    x->x_shiftdown = 0;
    outlet_new(&x->x_obj, &s_bang);
    return (x);
}

static void toggle_ff(t_toggle *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void toggle_setup(void)
{
    toggle_class = class_new(gensym("tgl"), (t_newmethod)toggle_new,
			     (t_method)toggle_ff, sizeof(t_toggle), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)toggle_new, gensym("toggle"), A_GIMME, 0);
    class_addbang(toggle_class, toggle_bang);
    class_addfloat(toggle_class, toggle_float);
    class_addlist(toggle_class, toggle_list);
    class_addmethod(toggle_class, (t_method)toggle_click, gensym("click"),
		    A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(toggle_class, (t_method)toggle_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_loadbang, gensym("loadbang"), 0);
    class_addmethod(toggle_class, (t_method)toggle_set, gensym("set"), A_FLOAT, 0);
    class_addmethod(toggle_class, (t_method)toggle_size, gensym("size"), A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_color, gensym("color"), A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(toggle_class, (t_method)toggle_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(toggle_class, (t_method)toggle_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(toggle_class, (t_method)toggle_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(toggle_class, (t_method)toggle_init, gensym("init"), A_FLOAT, 0);
    class_addmethod(toggle_class, (t_method)toggle_nonzero, gensym("nonzero"), A_FLOAT, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    toggle_widgetbehavior.w_getrectfn = toggle_getrect;
    toggle_widgetbehavior.w_displacefn = toggle_displace;
    toggle_widgetbehavior.w_selectfn = toggle_select;
    toggle_widgetbehavior.w_activatefn = NULL;
    toggle_widgetbehavior.w_deletefn = toggle_delete;
    toggle_widgetbehavior.w_visfn = toggle_vis;
#if (PD_VERSION_MINOR > 31)
    toggle_widgetbehavior.w_clickfn = toggle_newclick;
    toggle_widgetbehavior.w_propertiesfn = toggle_properties;
#endif
    toggle_widgetbehavior.w_savefn = toggle_save;
    class_setwidget(toggle_class, &toggle_widgetbehavior);
    class_sethelpsymbol(toggle_class, gensym("toggle"));
}

/* ---------- my_canvas  my gui-canvas for a window ---------------- */

t_widgetbehavior my_canvas_widgetbehavior;
static t_class *my_canvas_class;

typedef struct _my_canvas
{
     t_object x_obj;
     t_atom   x_at[3];
     t_glist  *x_glist;
     int      x_size;
     int      x_w;
     int      x_h;
     char     x_font[16];
     int      x_fontsize;
     int      x_ldx;
     int      x_ldy;
     int      x_bcol;
     int      x_lcol;
     int      x_unique;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_selected;
     int      x_shiftdown;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
} t_my_canvas;

/* widget helper functions */

void my_canvas_draw(t_my_canvas *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(firsttime)
    {
        sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -outline #%6.6x -tags %xRECT\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_w, x->x_obj.te_ypos + x->x_h,
		 x->x_bcol, x->x_bcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -outline #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size,
		 x->x_bcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
    }
    else
    {
        sys_vgui(".x%x.c coords %xRECT %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos, x->x_obj.te_xpos + x->x_w,
		 x->x_obj.te_ypos + x->x_h);
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx,
		     x->x_obj.te_ypos+x->x_ldy);
    }
}

void my_canvas_erase(t_my_canvas* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xRECT\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void my_canvas_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_my_canvas *x = (t_my_canvas *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void my_canvas_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_my_canvas *x = (t_my_canvas *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    my_canvas_draw(x, glist, 0);
}

static void my_canvas_select(t_gobj *z, t_glist *glist, int selected)
{
    t_my_canvas *x = (t_my_canvas *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, x->x_bcol);
    }
}

static void my_canvas_delete(t_gobj *z, t_glist *glist)
{
}

static void my_canvas_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_my_canvas *x = (t_my_canvas *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 my_canvas_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 my_canvas_erase(x, glist);
         rtext_free(y);
    }
}

static void my_canvas_save(t_gobj *z, t_binbuf *b)
{
    t_my_canvas *x = (t_my_canvas *)z;
    int bcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiisssiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("my_canvas"), x->x_size, x->x_w, x->x_h,
		x->x_snd, x->x_rcv, x->x_lab, x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8+x->x_unique, x->x_fontsize,
		bcol, lcol);
    binbuf_addv(b, ";");
}

static void my_canvas_properties(t_gobj *z, t_glist *owner)
{
    t_my_canvas *x = (t_my_canvas *)z;
    char buf[600];

    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s MY_CANVAS \
	    ------selectable_dimensions(pix):------ %g %g size: 0.0 0.0 empty \
	    ------visible_rectangle(pix)(pix):------ %g width: %g height: %g \
	    %g empty empty %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_size, 1.0,
	    (float)x->x_w, (float)x->x_h, 0.0,/*no_schedule*/
	    -1.0, -1.0, -1.0,/*no linlog, no init, no multi*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, -1/*no frontcolor*/, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void my_canvas_get_pos(t_my_canvas *x)
{
    if(x->x_snd_able && x->x_snd->s_thing)
    {
	x->x_at[0].a_w.w_float = x->x_obj.te_xpos;
	x->x_at[1].a_w.w_float = x->x_obj.te_ypos;
	pd_list(x->x_snd->s_thing, &s_list, 2, x->x_at);
    }
}

static void my_canvas_dialog(t_my_canvas *x, t_symbol *s, int argc, t_atom *argv)
{
    char font[20];
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int a = (int)atom_getintarg(0, argc, argv);
    int w = (int)atom_getintarg(2, argc, argv);
    int h = (int)atom_getintarg(3, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(snd))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }
    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    if(a < 1)
	a = 1;
    x->x_size = a;
    if(w < 1)
	w = 1;
    x->x_w = w;
    if(h < 1)
	h = 1;
    x->x_h = h;
    x->x_lcol = lcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    sys_vgui(".x%x.c itemconfigure %xRECT -fill #%6.6x -outline #%6.6x\n", canvas, x,
	     x->x_bcol, x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x,
	         x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_bcol);
    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize, x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize, x->x_lcol, x->x_lab->s_name);
    my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_size(t_my_canvas *x, t_float size)
{
    if(size < 1.0)
        size = 1.0;

    x->x_size = (int)size;
    if(glist_isvisible(x->x_glist))
        my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_delta(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
        my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_pos(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
        my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_vis_size(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    int i;

    i = (int)atom_getintarg(0, ac, av);
    if(i < 1)
	i = 1;
    x->x_w = i;
    i = (int)atom_getintarg(1, ac, av);
    if(i < 1)
	i = 1;
    x->x_h = i;
    if(glist_isvisible(x->x_glist))
        my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_color(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i,j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    {
	sys_vgui(".x%x.c itemconfigure %xRECT -fill #%6.6x -outline #%6.6x\n",
		 canvas, x, x->x_bcol, x->x_bcol);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_bcol);
	if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);
    }
}

static void my_canvas_send(t_my_canvas *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void my_canvas_receive(t_my_canvas *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void my_canvas_label(t_my_canvas *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    x->x_lab = s;
    if(glist_isvisible(x->x_glist))
    {
	if(!strcmp(s->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, s->s_name);
    }
}

static void my_canvas_label_pos(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
        sys_vgui(".x%x.c coords %xLABEL %d %d\n", glist_getcanvas(x->x_glist), x,
	     x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
}

static void my_canvas_label_font(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
        sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
}

static void my_canvas_list(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	t_symbol *key = atom_getsymbolarg(1, ac, av);
	int keydown = atom_getintarg(0, ac, av);

	if(keydown)
	{
	    int refresh = 1,i,d=1;
	    static char buf[20];

	    buf[0] = 0;
	    if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		x->x_shiftdown = 1;
	    else
	    {
		if(x->x_shiftdown)
		    d = 10;

		if(!strcmp(key->s_name, "Up"))
		    x->x_obj.te_ypos -= d;
		else if(!strcmp(key->s_name, "Down"))
		    x->x_obj.te_ypos += d;
		else if(!strcmp(key->s_name, "Left"))
		    x->x_obj.te_xpos -= d;
		else if(!strcmp(key->s_name, "Right"))
		    x->x_obj.te_xpos += d;
		else
		    refresh = 0;
		if(refresh)
		{
		    my_canvas_draw(x, x->x_glist, 0);
		}
	    }
	}
	else
	{
	    if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		x->x_shiftdown = 0;
	}
    }
    else
        post("error: my_canvas: no method for 'list'");
}

static void *my_canvas_new(t_symbol *s, int argc, t_atom *argv)
{
    t_my_canvas *x = (t_my_canvas *)pd_new(my_canvas_class);
    int bcol= -233017;
    int lcol= -66577;
    int a=IEM_GUI_DEFAULTSIZE, w=100, h=60;
    int ldx=20, ldy=12, f=2, fs=14;
    int sndable=1, rcvable=1, i=0, unique=0;
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if(((argc >= 10)&&(argc <= 12))
       &&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)&&IS_A_FLOAT(argv,2))
    {
	a = (int)atom_getintarg(0, argc, argv);
	w = (int)atom_getintarg(1, argc, argv);
	h = (int)atom_getintarg(2, argc, argv);
    }
    if((argc == 12)&&IS_A_SYMBOL(argv,3)&&IS_A_SYMBOL(argv,4))
    {
	i = 2;
	snd = atom_getsymbolarg(3, argc, argv);
	rcv = atom_getsymbolarg(4, argc, argv);
    }
    else if((argc == 11)&&IS_A_SYMBOL(argv,3))
    {
	i = 1;
	rcv = atom_getsymbolarg(3, argc, argv);
    }

    if(((argc >= 10)&&(argc <= 12))
       &&IS_A_SYMBOL(argv,i+3)&&IS_A_FLOAT(argv,i+4)
       &&IS_A_FLOAT(argv,i+5)&&IS_A_FLOAT(argv,i+6)
       &&IS_A_FLOAT(argv,i+7)&&IS_A_FLOAT(argv,i+8)
       &&IS_A_FLOAT(argv,i+9))
    {
	lab = atom_getsymbolarg(i+3, argc, argv);
	ldx = (int)atom_getintarg(i+4, argc, argv);
	ldy = (int)atom_getintarg(i+5, argc, argv);
        f = ((int)atom_getintarg(i+6, argc, argv))&255;
	unique = ((int)atom_getintarg(i+6, argc, argv))&1792;
	fs = (int)atom_getintarg(i+7, argc, argv);
	bcol = (int)atom_getintarg(i+8, argc, argv);
	lcol = (int)atom_getintarg(i+9, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(a < 1)
	a = 1;
    x->x_size = a;
    if(w < 1)
	w = 1;
    x->x_w = w;
    if(h < 1)
	h = 1;
    x->x_h = h;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    reconvert2unique(snd, unique & 256);
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
	fs = 4;
    x->x_fontsize = fs;
    if(lcol < 0)
    {
	lcol = -1 - lcol;
	x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
	bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    x->x_shiftdown = 0;
    x->x_at[0].a_type = A_FLOAT;
    x->x_at[1].a_type = A_FLOAT;
    return (x);
}

static void my_canvas_ff(t_my_canvas *x)
{
    if(x->x_selected)
	pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void my_canvas_setup(void)
{
    my_canvas_class = class_new(gensym("cnv"), (t_newmethod)my_canvas_new,
			     (t_method)my_canvas_ff, sizeof(t_my_canvas), CLASS_NOINLET, A_GIMME, 0);
    class_addcreator((t_newmethod)my_canvas_new, gensym("my_canvas"), A_GIMME, 0);
    class_addlist(my_canvas_class, my_canvas_list);
    class_addmethod(my_canvas_class, (t_method)my_canvas_dialog, gensym("dialog"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_size, gensym("size"), A_DEFFLOAT, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_vis_size, gensym("vis_size"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_color, gensym("color"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_get_pos, gensym("get_pos"), 0);

    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    my_canvas_widgetbehavior.w_getrectfn = my_canvas_getrect;
    my_canvas_widgetbehavior.w_displacefn = my_canvas_displace;
    my_canvas_widgetbehavior.w_selectfn = my_canvas_select;
    my_canvas_widgetbehavior.w_activatefn = NULL;
    my_canvas_widgetbehavior.w_deletefn = my_canvas_delete;
    my_canvas_widgetbehavior.w_visfn = my_canvas_vis;
#if (PD_VERSION_MINOR > 31)
    my_canvas_widgetbehavior.w_clickfn = NULL;
    my_canvas_widgetbehavior.w_propertiesfn = my_canvas_properties;
#endif
    my_canvas_widgetbehavior.w_savefn = my_canvas_save;
    class_setwidget(my_canvas_class, &my_canvas_widgetbehavior);
    class_sethelpsymbol(my_canvas_class, gensym("my_canvas"));
}

/* ------------ vertical  slider ----------------------- */

typedef struct _vslider
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_w;
     int      x_h;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     int      x_pos;
     int      x_val;
     int      x_lcol;
     int      x_fcol;
     int      x_bcol;
     int      x_unique;
     int      x_lin0_log1;
     int      x_loadinit;
     int      x_steady;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_finemoved;
     int      x_selected;
     int      x_shiftdown;
     double   x_min;
     double   x_max;
     double   x_k;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
} t_vslider;

t_widgetbehavior vslider_widgetbehavior;
static t_class *vslider_class;

/* widget helper functions */

static void vslider_update(t_vslider *x, t_glist *glist)
{
    if (glist_isvisible(glist))
    {
	int r = x->x_obj.te_ypos + x->x_h - (x->x_val + 50)/100;
	sys_vgui(".x%x.c coords %xKNOB %d %d %d %d\n",
		 glist_getcanvas(glist), x, x->x_obj.te_xpos+1, r,
		 x->x_obj.te_xpos + x->x_w, r);
    }
}

static void vslider_draw(t_vslider *x, t_glist *glist, int firsttime)
{
    int r = x->x_obj.te_ypos + x->x_h - (x->x_val + 50)/100;
    t_canvas *canvas=glist_getcanvas(glist);

    if (firsttime)
    {
	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + x->x_w, x->x_obj.te_ypos + x->x_h+3,
		 x->x_bcol, x);
	sys_vgui(".x%x.c create line %d %d %d %d -width 3 -fill #%6.6x -tags %xKNOB\n",
		 canvas, x->x_obj.te_xpos+1, r,
		 x->x_obj.te_xpos + x->x_w, r, x->x_fcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas,
		 x->x_obj.te_xpos, x->x_obj.te_ypos + x->x_h+2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos + x->x_h+3,
		 x, 0);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas,
		 x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos-1,
		 x, 0);
    }
    else
    {
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x,
		 x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + x->x_w, x->x_obj.te_ypos + x->x_h+3);
        sys_vgui(".x%x.c coords %xKNOB %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos+1, r,
		 x->x_obj.te_xpos + x->x_w, r);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
	sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos, x->x_obj.te_ypos + x->x_h+2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos + x->x_h+3);
	sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos-1);
    }
}

static void vslider_erase(t_vslider* x,t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xKNOB\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
}

/* ------------------------ hsld widgetbehaviour----------------------------- */


static void vslider_getrect(t_gobj *z, t_glist *owner,
    int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_vslider* x = (t_vslider*)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos - 2;
    *xp2 = x->x_obj.te_xpos + x->x_w;
    *yp2 = x->x_obj.te_ypos + x->x_h + 3;
}

static void vslider_displace(t_gobj *z, t_glist *glist,
    int dx, int dy)
{
    t_vslider* x = (t_vslider*)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    vslider_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text*)z);
}

static void vslider_select(t_gobj *z, t_glist *glist, int selected)
{
    t_vslider *x = (t_vslider *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void vslider_delete(t_gobj *z, t_glist *glist)
{
    t_vslider *x = (t_vslider *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void vslider_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_vslider *x = (t_vslider *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 vslider_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 vslider_erase(x, glist);
         rtext_free(y);
    }
}

static void vslider_save(t_gobj *z, t_binbuf *b)
{
    t_vslider *x = (t_vslider *)z;
    int bcol, fcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    fcol = -1 - (((0xfc0000 & x->x_fcol) >> 6)|((0xfc00 & x->x_fcol) >> 4)|((0xfc & x->x_fcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiffiisssiiiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("vslider"), x->x_w, x->x_h,
		(float)x->x_min, (float)x->x_max,
                x->x_lin0_log1, x->x_loadinit,
		x->x_snd, x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8+x->x_unique, x->x_fontsize,
		bcol, fcol, lcol,
	        x->x_val, x->x_steady);
    binbuf_addv(b, ";");
}

void vslider_check_height(t_vslider *x, int h)
{
    if(h < IEM_SL_MINSIZE)
	h = IEM_SL_MINSIZE;
    x->x_h = h;
    if(x->x_val > (x->x_h*100 - 100))
    {
	x->x_pos = x->x_h*100 - 100;
	x->x_val = x->x_pos;
    }
    if(x->x_lin0_log1)
	x->x_k = log(x->x_max/x->x_min)/(double)(x->x_h - 1);
    else
	x->x_k = (x->x_max - x->x_min)/(double)(x->x_h - 1);
}

void vslider_check_minmax(t_vslider *x, double min, double max)
{
    if(x->x_lin0_log1)
    {
	if((min == 0.0)&&(max == 0.0))
            max = 1.0;
	if(max > 0.0)
	{
	    if(min <= 0.0)
                min = 0.01*max;
	}
	else
	{
            if(min > 0.0)
                max = 0.01*min;
	}
    }
    x->x_min = min;
    x->x_max = max;
    if(x->x_lin0_log1)
        x->x_k = log(x->x_max/x->x_min)/(double)(x->x_h - 1);
    else
	x->x_k = (x->x_max - x->x_min)/(double)(x->x_h - 1);
}

static void vslider_properties(t_gobj *z, t_glist *owner)
{
    t_vslider *x = (t_vslider *)z;
    char buf[600];
    int steady=-23;

    if(x->x_steady)
        steady = -22;
    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s VSLIDER \
	    --------dimensions(pix)(pix):-------- %g %g width: %g %g height: \
	    -----------output-range:----------- %g bottom: %g top: %g \
	    %g lin log %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_w, (float)IEM_GUI_MINSIZE, (float)x->x_h, (float)IEM_SL_MINSIZE,
	    x->x_min, x->x_max, 0.0,/*no_schedule*/
	    (float)x->x_lin0_log1, (float)x->x_loadinit, (float)steady,/*no multi, but iem-characteristic*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, 0xffffff & x->x_fcol, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void vslider_bang(t_vslider *x)
{
    double out;

    if(x->x_lin0_log1)
	out = x->x_min*exp(x->x_k*(double)(x->x_val)*0.01);
    else
	out = (double)(x->x_val)*0.01*x->x_k + x->x_min;
    if((out < 1.0e-10)&&(out > -1.0e-10))
        out = 0.0;

    outlet_float(x->x_obj.ob_outlet, out);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_float(x->x_snd->s_thing, out);
}

static void vslider_dialog(t_vslider *x, t_symbol *s, int argc, t_atom *argv)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int w = (int)atom_getintarg(0, argc, argv);
    int h = (int)atom_getintarg(1, argc, argv);
    double min = (double)atom_getfloatarg(2, argc, argv);
    double max = (double)atom_getfloatarg(3, argc, argv);
    int lilo = (int)atom_getintarg(4, argc, argv);
    int init = (int)atom_getintarg(5, argc, argv);
    int steady = (int)atom_getintarg(6, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int fcol = (int)atom_getintarg(15, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    if(lilo != 0) lilo = 1;
    x->x_lin0_log1 = lilo;
    if(init != 0) init = 1;
    x->x_loadinit = init;
    if(steady == 3)
	x->x_steady = 0;
    else
        x->x_steady = 1;
    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(snd))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }

    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_w = iem_gui_clip_size(w);
    vslider_check_height(x, h);
    vslider_check_minmax(x, min, max);
    x->x_lcol = lcol & 0xffffff;
    x->x_fcol = fcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);
    sys_vgui(".x%x.c itemconfigure %xKNOB -fill #%6.6x\n", canvas,
	     x, x->x_fcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
	     x, x->x_bcol);
    vslider_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(canvas, (t_text*)x);
}

static void vslider_motion(t_vslider *x, t_floatarg dx, t_floatarg dy)
{
    int old = x->x_val;

    if(x->x_finemoved)
        x->x_pos -= (int)dy;
    else
	x->x_pos -= 100*(int)dy;
    x->x_val = x->x_pos;
    if(x->x_val > (100*x->x_h - 100))
    {
	x->x_val = 100*x->x_h - 100;
        x->x_pos += 50;
	/*x->x_pos /= 100;
	 x->x_pos *= 100;*/
        x->x_pos -= x->x_pos%100;
    }
    if(x->x_val < 0)
    {
	x->x_val = 0;
	x->x_pos -= 50;
	/*x->x_pos /= 100;
	 x->x_pos *= 100;*/
        x->x_pos -= x->x_pos%100;
    }
    if(old != x->x_val)
    {
	vslider_update(x, x->x_glist);
	vslider_bang(x);
    }
}

static void vslider_click(t_vslider *x, t_floatarg xpos, t_floatarg ypos,
		       t_floatarg shift, t_floatarg ctrl, t_floatarg alt)
{
    if(!x->x_steady)
	x->x_val = (int)(100.0 * (x->x_h + x->x_obj.te_ypos - ypos));
    if(x->x_val > (100*x->x_h - 100))
	x->x_val = 100*x->x_h - 100;
    if(x->x_val < 0)
	x->x_val = 0;
    x->x_pos = x->x_val;
    vslider_update(x, x->x_glist);
    vslider_bang(x);
#if (PD_VERSION_MINOR > 31)
    glist_grab(x->x_glist, &x->x_obj.te_g, (t_glistmotionfn)vslider_motion,
	       0, xpos, ypos);
#else
    glist_grab(x->x_glist, &x->x_obj.te_g, xpos, ypos);
#endif
}

#if (PD_VERSION_MINOR > 31)
static int vslider_newclick(t_gobj *z, struct _glist *glist,
    int xpix, int ypix, int shift, int alt, int dbl, int doit)
{
    t_vslider* x = (t_vslider *)z;

    if(doit)
    {
	vslider_click( x, (t_floatarg)xpix, (t_floatarg)ypix, (t_floatarg)shift,
		    0, (t_floatarg)alt);
	if(shift)
	    x->x_finemoved = 1;
	else
            x->x_finemoved = 0;
    }
    return (1);
}
#endif

static void vslider_set(t_vslider *x, t_floatarg f)
{
    double g;

    if(f > x->x_max)
	f = x->x_max;
    if(f < x->x_min)
	f = x->x_min;
    if(x->x_lin0_log1)
	g = log(f/x->x_min)/x->x_k;
    else
	g = (f - x->x_min) / x->x_k;
    x->x_val = (int)(100.0*g + 0.49999);
    x->x_pos = x->x_val;
    vslider_update(x, x->x_glist);
}

static void vslider_size(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_w = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    vslider_check_height(x, (int)atom_getintarg(1, ac, av));
    if(glist_isvisible(x->x_glist))
    {
        vslider_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void vslider_delta(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        vslider_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void vslider_pos(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
        vslider_draw(x, x->x_glist, 0);
        canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void vslider_range(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    vslider_check_minmax(x, (double)atom_getfloatarg(0, ac, av),
			 (double)atom_getfloatarg(1, ac, av));
}

static void vslider_color(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i,j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_fcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_fcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(2, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    {
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas,
	     x, x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
        sys_vgui(".x%x.c itemconfigure %xKNOB -fill #%6.6x\n", canvas,
	     x, x->x_fcol);
        sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
		 x, x->x_bcol);
    }
}

static void vslider_send(t_vslider *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 256;
    else
        x->x_unique &= 512+1024;
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void vslider_receive(t_vslider *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
     else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void vslider_label(t_vslider *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    x->x_lab = s;
    if(glist_isvisible(x->x_glist))
    {
        if(!strcmp(s->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, "");
        else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		     glist_getcanvas(x->x_glist), x, s->s_name);
    }
}

static void vslider_label_pos(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
        sys_vgui(".x%x.c coords %xLABEL %d %d\n", glist_getcanvas(x->x_glist), x,
	     x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
}

static void vslider_label_font(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
        sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
}

static void vslider_log(t_vslider *x)
{
    x->x_lin0_log1 = 1;
    vslider_check_minmax(x, x->x_min, x->x_max);
}

static void vslider_lin(t_vslider *x)
{
    x->x_lin0_log1 = 0;
    x->x_k = (x->x_max - x->x_min)/(double)(x->x_h - 1);
}

static void vslider_init(t_vslider *x, t_floatarg f)
{
    x->x_loadinit = (f==0.0)?0:1;
}

static void vslider_steady(t_vslider *x, t_floatarg f)
{
    x->x_steady = (f==0.0)?0:1;
}

static void vslider_float(t_vslider *x, t_floatarg f)
{
    vslider_set(x, f);
    vslider_bang(x);
}

static void vslider_loadbang(t_vslider *x)
{
    if(!sys_noloadbang && x->x_loadinit)
    {
	vslider_update(x, x->x_glist);
	vslider_bang(x);
    }
}

static void vslider_list(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto vsliderlistelse;
	else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;
		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
                    {
			vslider_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
	return;
    }
vsliderlistelse:
    if(IS_A_FLOAT(av,0))
	vslider_float(x, atom_getfloatarg(0, ac, av));
}

static void *vslider_new(t_symbol *s, int argc, t_atom *argv)
{
    t_vslider *x = (t_vslider *)pd_new(vslider_class);
    int bcol= -262144;
    int fcol= -1;
    int lcol= -1;
    int w=IEM_GUI_DEFAULTSIZE, h=IEM_SL_DEFAULTSIZE;
    int lilo=0, init=0, ldx=20, ldy=8, f=0, fs=8;
    int sndable=1, rcvable=1, v=0, steady=1, unique=0;
    double min=0.0, max=(double)(IEM_SL_DEFAULTSIZE-1);
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if(((argc == 17)||(argc == 18))&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)
       &&IS_A_FLOAT(argv,4)&&IS_A_FLOAT(argv,5)
       &&IS_A_SYMBOL(argv,6)&&IS_A_SYMBOL(argv,7)
       &&IS_A_SYMBOL(argv,8)&&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10)
       &&IS_A_FLOAT(argv,11)&&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13)
       &&IS_A_FLOAT(argv,14)&&IS_A_FLOAT(argv,15)&&IS_A_FLOAT(argv,16))
    {
        w = (int)atom_getintarg(0, argc, argv);
	h = (int)atom_getintarg(1, argc, argv);
	min = (double)atom_getfloatarg(2, argc, argv);
	max = (double)atom_getfloatarg(3, argc, argv);
	lilo = (int)atom_getintarg(4, argc, argv);
	init = (int)atom_getintarg(5, argc, argv);
        snd = atom_getsymbolarg(6, argc, argv);
	rcv = atom_getsymbolarg(7, argc, argv);
	lab = atom_getsymbolarg(8, argc, argv);
	ldx = (int)atom_getintarg(9, argc, argv);
	ldy = (int)atom_getintarg(10, argc, argv);
        f = ((int)atom_getintarg(11, argc, argv))&255;
	unique = ((int)atom_getintarg(11, argc, argv))&1792;
        fs = (int)atom_getintarg(12, argc, argv);
        bcol = (int)atom_getintarg(13, argc, argv);
	fcol = (int)atom_getintarg(14, argc, argv);
        lcol = (int)atom_getintarg(15, argc, argv);
	v = (int)atom_getintarg(16, argc, argv);
    }
    if((argc == 18)&&IS_A_FLOAT(argv,17))
	steady = (int)atom_getintarg(17, argc, argv);
    if(init != 0) init = 1;
    x->x_loadinit = init;
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(x->x_loadinit)
	x->x_val = v;
    else
        x->x_val = 0;
    x->x_pos = x->x_val;
    if(lilo != 0) lilo = 1;
    x->x_lin0_log1 = lilo;
    if(steady != 0) steady = 1;
    x->x_steady = steady;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    reconvert2unique(snd, unique & 256);
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_w = iem_gui_clip_size(w);
    vslider_check_height(x, h);
    vslider_check_minmax(x, min, max);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(fcol < 0)
    {
        fcol = -1 - fcol;
	x->x_fcol = ((fcol & 0x3f000) << 6)|((fcol & 0xfc0) << 4)|((fcol & 0x3f) << 2);
    }
    else
    {
	fcol = iem_gui_modulo_color(fcol);
	x->x_fcol = iem_gui_color_hex[fcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_finemoved = 0;
    x->x_selected = 0;
    x->x_shiftdown = 0;
    outlet_new(&x->x_obj, &s_float);
    return (x);
}

static void vslider_free(t_vslider *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void vslider_setup(void)
{
    vslider_class = class_new(gensym("vsl"), (t_newmethod)vslider_new,
			      (t_method)vslider_free, sizeof(t_vslider), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)vslider_new, gensym("vslider"), A_GIMME, 0);
    class_addbang(vslider_class,vslider_bang);
    class_addfloat(vslider_class,vslider_float);
    class_addlist(vslider_class, vslider_list);
    class_addmethod(vslider_class, (t_method)vslider_click, gensym("click"),
		    A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_motion, gensym("motion"),
		    A_FLOAT, A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_loadbang, gensym("loadbang"), 0);
    class_addmethod(vslider_class, (t_method)vslider_set, gensym("set"), A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_size, gensym("size"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_range, gensym("range"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_color, gensym("color"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(vslider_class, (t_method)vslider_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(vslider_class, (t_method)vslider_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(vslider_class, (t_method)vslider_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_log, gensym("log"), 0);
    class_addmethod(vslider_class, (t_method)vslider_lin, gensym("lin"), 0);
    class_addmethod(vslider_class, (t_method)vslider_init, gensym("init"), A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_steady, gensym("steady"), A_FLOAT, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    vslider_widgetbehavior.w_getrectfn =    vslider_getrect;
    vslider_widgetbehavior.w_displacefn =   vslider_displace;
    vslider_widgetbehavior.w_selectfn =     vslider_select;
    vslider_widgetbehavior.w_activatefn =   NULL;
    vslider_widgetbehavior.w_deletefn =     vslider_delete;
    vslider_widgetbehavior.w_visfn =        vslider_vis;
#if (PD_VERSION_MINOR > 31)
    vslider_widgetbehavior.w_clickfn =      vslider_newclick;
    vslider_widgetbehavior.w_propertiesfn = vslider_properties;;
#endif
    vslider_widgetbehavior.w_savefn =       vslider_save;
    class_setwidget(vslider_class, &vslider_widgetbehavior);
    class_sethelpsymbol(vslider_class, gensym("vslider"));
}

/* ----- peak- & rms- vu-meter-display ---------- */

int iem_gui_vu_db2i[]=
{
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
     9, 9, 9, 9, 9,10,10,10,10,10,
    11,11,11,11,11,12,12,12,12,12,
    13,13,13,13,14,14,14,14,15,15,
    15,15,16,16,16,16,17,17,17,18,
    18,18,19,19,19,20,20,20,21,21,
    22,22,23,23,24,24,25,26,27,28,
    29,30,31,32,33,33,34,34,35,35,
    36,36,37,37,37,38,38,38,39,39,
    39,39,39,39,40,40
};

int iem_gui_vu_col[]=
{
    0,17,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
    15,15,15,15,15,15,15,15,15,15,14,14,13,13,13,13,13,13,13,13,13,13,13,19,19,19
};

char *iem_gui_vu_scale_str[]=
{
    "",
    "<-99",
    "",
    "",
    "",
    "-50",
    "",
    "",
    "",
    "-30",
    "",
    "",
    "",
    "-20",
    "",
    "",
    "",
    "-12",
    "",
    "",
    "",
    "-6",
    "",
    "",
    "",
    "-2",
    "",
    "",
    "",
    "-0dB",
    "",
    "",
    "",
    "+2",
    "",
    "",
    "",
    "+6",
    "",
    "",
    "",
    ">+12",
    "",
    "",
    "",
    "",
    "",
};

typedef struct _vu
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_w;
     int      x_h;
     int      x_led_size;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     int      x_peak;
     int      x_rms;
     float    x_fp;
     float    x_fr;
     int      x_lcol;
     int      x_bcol;
     int      x_unique;
     int      x_rcv_able;
     int      x_scale;
     int      x_selected;
     int      x_shiftdown;
     t_symbol *x_rcv;
     t_symbol *x_lab;
     void     *x_out_rms;
     void     *x_out_peak;
} t_vu;

t_widgetbehavior vu_widgetbehavior;
static t_class *vu_class;

/* widget helper functions */

static void vu_update_rms(t_vu *x, t_glist *glist)
{
    if(glist_isvisible(glist))
    {
	int w4=x->x_w/4, off=x->x_obj.te_ypos-1;
	int quad1=x->x_obj.te_xpos+w4+1, quad3=x->x_obj.te_xpos+x->x_w-w4-1;

	sys_vgui(".x%x.c coords %xRCOVER %d %d %d %d\n",
		 glist_getcanvas(glist), x, quad1, off, quad3,
		 off + (x->x_led_size+1)*(IEM_VU_STEPS-x->x_rms));
    }
}

static void vu_update_peak(t_vu *x, t_glist *glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(glist_isvisible(glist))
    {
	if(x->x_peak)
	{
	    int i=iem_gui_vu_col[x->x_peak];
	    int j=x->x_obj.te_ypos + (x->x_led_size+1)*(IEM_VU_STEPS+1-x->x_peak)
		- (x->x_led_size+1)/2;

            sys_vgui(".x%x.c coords %xPLED %d %d %d %d\n", canvas, x,
		     x->x_obj.te_xpos, j,
		     x->x_obj.te_xpos+x->x_w+1, j);
	    sys_vgui(".x%x.c itemconfigure %xPLED -fill #%6.6x\n", canvas, x,
		     iem_gui_color_hex[i]);
	}
	else
	{
	    int mid=x->x_obj.te_xpos+x->x_w/2;

            sys_vgui(".x%x.c itemconfigure %xPLED -fill #%6.6x\n",
		     canvas, x, x->x_bcol);
            sys_vgui(".x%x.c coords %xPLED %d %d %d %d\n",
		 canvas, x, mid, x->x_obj.te_ypos+20,
		     mid, x->x_obj.te_ypos+20);
	}
    }
}

static void vu_draw(t_vu *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if (firsttime)
    {
	int w4=x->x_w/4, mid=x->x_obj.te_xpos+x->x_w/2, quad1=x->x_obj.te_xpos+w4+1;
	int quad3=x->x_obj.te_xpos+x->x_w-w4, end=x->x_obj.te_xpos+x->x_w+4;
	int k1=x->x_led_size+1, k2=IEM_VU_STEPS+1, k3=k1/2;
	int led_col, yyy, i, k4=x->x_obj.te_ypos-k3;

	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1,x->x_obj.te_ypos+x->x_h+2, x->x_bcol, x);
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    led_col = iem_gui_vu_col[i];
            yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c create line %d %d %d %d -width %d -fill #%6.6x -tags %xRLED%d\n",
		     canvas, quad1, yyy, quad3, yyy, x->x_led_size, iem_gui_color_hex[led_col], x, i);
	    if(((i+2)&3) && (x->x_scale))
		sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
			 -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
			 canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
			 x->x_lcol, x, i);
	}
	if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
		     canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		     x->x_lcol, x, i);
	}
        sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -outline #%6.6x -tags %xRCOVER\n",
		 canvas, quad1, x->x_obj.te_ypos-1, quad3-1,
		 x->x_obj.te_ypos-1 + k1*IEM_VU_STEPS, x->x_bcol, x->x_bcol, x);
        sys_vgui(".x%x.c create line %d %d %d %d -width %d -fill #%6.6x -tags %xPLED\n",
		 canvas, mid, x->x_obj.te_ypos+10,
		     mid, x->x_obj.te_ypos+10, x->x_led_size, x->x_bcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize, x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize, x->x_lcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos + x->x_h+2,
		 x, 0);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos + x->x_h+2,
		 x, 1);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos-1,
		 x, 0);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos-1,
		 x, 1);
    }
    else
    {
	int w4=x->x_w/4, quad1=x->x_obj.te_xpos+w4+1;
	int quad3=x->x_obj.te_xpos+x->x_w-w4, end=x->x_obj.te_xpos+x->x_w+4;
	int k1=x->x_led_size+1, k2=IEM_VU_STEPS+1, k3=k1/2;
	int yyy, i, k4=x->x_obj.te_ypos-k3;

        sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1,x->x_obj.te_ypos+x->x_h+2);
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
            yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c coords %xRLED%d %d %d %d %d\n",
		     canvas, x, i, quad1, yyy, quad3, yyy);
	    if(((i+2)&3) && (x->x_scale))
                sys_vgui(".x%x.c coords %xSCALE%d %d %d\n",
		     canvas, x, i, end, yyy+k3);
	}
	if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c coords %xSCALE%d %d %d\n",
		 canvas, x, i, end, yyy+k3);
	}
        vu_update_peak(x, glist);
	vu_update_rms(x, glist);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx,
		 x->x_obj.te_ypos+x->x_ldy);
	sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos + x->x_h+2);
        sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 1,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos + x->x_h+2);
	sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos-1);
        sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 1,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos-1);
    }
}

static void vu_confitem(t_vu* x,t_glist* glist)
{
    int i;
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x, x->x_bcol);
    for(i=1; i<=IEM_VU_STEPS; i++)
    {
        sys_vgui(".x%x.c itemconfigure %xRLED%d -width %d\n", canvas, x, i,
		     x->x_led_size);
	if(((i+2)&3) && (x->x_scale))
	    sys_vgui(".x%x.c itemconfigure %xSCALE%d -text {%s} -font {%s %d bold} -fill #%6.6x\n",
		     canvas, x, i, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
    }
    if(x->x_scale)
    {
	i=IEM_VU_STEPS+1;
	sys_vgui(".x%x.c itemconfigure %xSCALE%d -text {%s} -font {%s %d bold} -fill #%6.6x\n",
		     canvas, x, i, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
    }
    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);

    sys_vgui(".x%x.c itemconfigure %xRCOVER -fill #%6.6x -outline #%6.6x\n", canvas,
	     x, x->x_bcol, x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xPLED -width %d\n", canvas, x,
		     x->x_led_size);
}

static void vu_erase(t_vu* x,t_glist* glist)
{
    int i;
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    for(i=1; i<=IEM_VU_STEPS; i++)
    {
	sys_vgui(".x%x.c delete %xRLED%d\n", canvas, x, i);
        if(((i+2)&3) && (x->x_scale))
            sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
    }
    if(x->x_scale)
    {
	    i=IEM_VU_STEPS+1;
	    sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
    }
    sys_vgui(".x%x.c delete %xPLED\n", canvas, x);
    sys_vgui(".x%x.c delete %xRCOVER\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 1);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 1);
}

/* ------------------------ vu widgetbehaviour----------------------------- */


static void vu_getrect(t_gobj *z, t_glist *owner,
    int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_vu* x = (t_vu*)z;

    *xp1 = x->x_obj.te_xpos - 1;
    *yp1 = x->x_obj.te_ypos - 2;
    *xp2 = x->x_obj.te_xpos + x->x_w + 1;
    *yp2 = x->x_obj.te_ypos + x->x_h + 2;
}

static void vu_displace(t_gobj *z, t_glist *glist,
    int dx, int dy)
{
    t_vu* x = (t_vu*)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    vu_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void vu_select(t_gobj *z, t_glist *glist, int selected)
{
    t_vu *x = (t_vu *)z;
    int i;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    if(((i+2)&3) && (x->x_scale))
		sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
			 canvas, x, i, IEM_GUI_COLOR_SELECTED);
	}
	if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
		 canvas, x, i, IEM_GUI_COLOR_SELECTED);
	}
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        for(i=1; i<=IEM_VU_STEPS; i++)
	{
            if(((i+2)&3) && (x->x_scale))
		sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
			 canvas, x, i, x->x_lcol);
	}
        if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
		  canvas, x, i, x->x_lcol);
	}
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void vu_delete(t_gobj *z, t_glist *glist)
{
    t_vu *x = (t_vu *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void vu_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_vu *x = (t_vu *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 vu_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 vu_erase(x, glist);
         rtext_free(y);
    }
}

static void vu_save(t_gobj *z, t_binbuf *b)
{
    t_vu *x = (t_vu *)z;
    int bcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiissiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("vu"), x->x_w, x->x_h,
                x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8+x->x_unique, x->x_fontsize,
		bcol, lcol, x->x_scale);
    binbuf_addv(b, ";");
}

void vu_check_height(t_vu *x, int h)
{
    int n;

    n = h / IEM_VU_STEPS;
    if(n < IEM_VU_MINSIZE)
	n = IEM_VU_MINSIZE;
    x->x_led_size = n-1;
    x->x_h = IEM_VU_STEPS * n;
}

static void vu_scale(t_vu *x, t_floatarg fscale)
{
    int i, scale = (int)fscale;

    if(scale != 0) scale = 1;
    if(x->x_scale && !scale)
    {
	t_canvas *canvas=glist_getcanvas(x->x_glist);

	x->x_scale = (int)scale;
        if(glist_isvisible(x->x_glist))
        {
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    if((i+2)&3)
		sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
	}
	i=IEM_VU_STEPS+1;
	sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
	}
    }
    if(!x->x_scale && scale)
    {
	int w4=x->x_w/4, end=x->x_obj.te_xpos+x->x_w+4;
	int k1=x->x_led_size+1, k2=IEM_VU_STEPS+1, k3=k1/2;
	int yyy, k4=x->x_obj.te_ypos-k3;
	t_canvas *canvas=glist_getcanvas(x->x_glist);

	x->x_scale = (int)scale;
        if(glist_isvisible(x->x_glist))
        {
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    yyy = k4 + k1*(k2-i);
	    if((i+2)&3)
		sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
			 -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
			 canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
			 x->x_lcol, x, i);
	}
	i=IEM_VU_STEPS+1;
	yyy = k4 + k1*(k2-i);
	sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		 -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
		 canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		 x->x_lcol, x, i);
	}
    }
}

static void vu_properties(t_gobj *z, t_glist *owner)
{
    t_vu *x = (t_vu *)z;
    char buf[600];
    
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s VU-METER \
	    --------dimensions(pix)(pix):-------- %g %g width: %g %g height: \
	    empty 0.0 empty 0.0 empty %g \
	    %g no_scale scale %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_w, (float)IEM_GUI_MINSIZE, (float)x->x_h, (float)(IEM_VU_STEPS*IEM_VU_MINSIZE),
	    0.0,/*no_schedule*/
	    (float)x->x_scale, -1.0, -1.0,/*no linlog, no init, no multi*/
	    "nosndno", x->x_rcv->s_name,/*no send*/
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, -1/*no front-color*/, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void vu_dialog(t_vu *x, t_symbol *s, int argc, t_atom *argv)
{
    int w = (int)atom_getintarg(0, argc, argv);
    int h = (int)atom_getintarg(1, argc, argv);
    int scale = (int)atom_getintarg(4, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int rcvable=1, i;

    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(sym_is_dollarunique(rcv))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(sym_is_dollarunique(lab))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = rcv;
    }
    x->x_rcv_able = rcvable;
    x->x_w = iem_gui_clip_size(w);
    x->x_lcol = lcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    vu_check_height(x, h);
    if(scale != 0)
	scale = 1;
    vu_scale(x, (float)scale);
    vu_draw(x, x->x_glist, 0);
    vu_confitem(x, x->x_glist);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vu_size(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_w = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    vu_check_height(x, (int)atom_getintarg(1, ac, av));
    if(glist_isvisible(x->x_glist))
    {
    vu_draw(x, x->x_glist, 0);
    vu_confitem(x, x->x_glist);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void vu_delta(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
    vu_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void vu_pos(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    {
    vu_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
    }
}

static void vu_color(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    int i,j;

    i = (int)atom_getintarg(0, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_bcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_bcol = (-1 -i)&0xffffff;
    i = (int)atom_getintarg(1, ac, av);
    if(i >= 0)
    {
        j = iem_gui_modulo_color(i);
	x->x_lcol = iem_gui_color_hex[(j)];
    }
    else
	x->x_lcol = (-1 -i)&0xffffff;
    if(glist_isvisible(x->x_glist))
    vu_confitem(x, x->x_glist);
}

static void vu_receive(t_vu *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 512;
    else
        x->x_unique &= 256+1024;
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    else if(!rcvable && x->x_rcv_able)
    {
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
        x->x_rcv = s;
    }
    x->x_rcv_able = rcvable;
}

static void vu_label(t_vu *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    if(sym_is_dollarunique(s))
	x->x_unique |= 1024;
    else
        x->x_unique &= 256+512;
    x->x_lab = s;
    if(glist_isvisible(x->x_glist))
    vu_confitem(x, x->x_glist);
}

static void vu_label_pos(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    if(glist_isvisible(x->x_glist))
    sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 glist_getcanvas(x->x_glist), x, x->x_obj.te_xpos+x->x_ldx,
	     x->x_obj.te_ypos+x->x_ldy);
}

static void vu_label_font(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i, f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    if(glist_isvisible(x->x_glist))
    vu_confitem(x, x->x_glist);
}

static void vu_float(t_vu *x, t_floatarg rms)
{
    int i;
    
    if(rms <= IEM_VU_MINDB)
	x->x_rms = 0;
    else if(rms >= IEM_VU_MAXDB)
	x->x_rms = IEM_VU_STEPS;
    else
    {
	int i = (int)(2.0*(rms + IEM_VU_OFFSET));
	x->x_rms = iem_gui_vu_db2i[i];
    }
    i = (int)(100.0*rms + 10000.5);
    rms = 0.01*(float)(i - 10000);
    x->x_fr = rms;
    outlet_float(x->x_out_rms, rms);
    vu_update_rms(x, x->x_glist);
}

static void vu_ft1(t_vu *x, t_floatarg peak)
{
    int i;

    if(peak <= IEM_VU_MINDB)
	x->x_peak = 0;
    else if(peak >= IEM_VU_MAXDB)
	x->x_peak = IEM_VU_STEPS;
    else
    {
	int i = (int)(2.0*(peak + IEM_VU_OFFSET));
	x->x_peak = iem_gui_vu_db2i[i];
    }
    i = (int)(100.0*peak + 10000.5);
    peak = 0.01*(float)(i - 10000);
    x->x_fp = peak;
    outlet_float(x->x_out_peak, peak);
    vu_update_peak(x, x->x_glist);
}

static void vu_list(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto vulistelse;
        else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;

		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
		    {
			vu_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
        return;
    }
vulistelse:
    if((IS_A_FLOAT(av,0))&&(IS_A_FLOAT(av,1)))
	{
	    vu_ft1(x, atom_getfloatarg(1, ac, av));
	    vu_float(x, atom_getfloatarg(0, ac, av));
	}
}

static void vu_bang(t_vu *x)
{
    outlet_float(x->x_out_peak, x->x_fp);
    outlet_float(x->x_out_rms, x->x_fr);
    vu_update_rms(x, x->x_glist);
    vu_update_peak(x, x->x_glist);
}

static void *vu_new(t_symbol *s, int argc, t_atom *argv)
{
    t_vu *x = (t_vu *)pd_new(vu_class);
    int bcol= -66577;
    int lcol= -1;
    int w=IEM_GUI_DEFAULTSIZE, h=IEM_VU_STEPS*IEM_VU_DEFAULTSIZE;
    int ldx=35, ldy=8, f=0, fs=8, scale=1;
    int rcvable=1, unique=0;
    t_symbol *rcv = gensym("empty"), *lab = gensym("empty");

    if((argc == 11)&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_SYMBOL(argv,2)&&IS_A_SYMBOL(argv,3)
       &&IS_A_FLOAT(argv,4)&&IS_A_FLOAT(argv,5)
       &&IS_A_FLOAT(argv,6)&&IS_A_FLOAT(argv,7)
       &&IS_A_FLOAT(argv,8)&&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10))
    {
        w = (int)atom_getintarg(0, argc, argv);
	h = (int)atom_getintarg(1, argc, argv);
	rcv = atom_getsymbolarg(2, argc, argv);
	lab = atom_getsymbolarg(3, argc, argv);
	ldx = (int)atom_getintarg(4, argc, argv);
	ldy = (int)atom_getintarg(5, argc, argv);
        f = ((int)atom_getintarg(6, argc, argv))&255;
	unique = ((int)atom_getintarg(6, argc, argv))&1792;
        fs = (int)atom_getintarg(7, argc, argv);
        bcol = (int)atom_getintarg(8, argc, argv);
	lcol = (int)atom_getintarg(9, argc, argv);
        scale = (int)atom_getintarg(10, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_rcv_able = rcvable;
    reconvert2unique(rcv, unique & 512);
    reconvert2unique(lab, unique & 1024);
    x->x_unique = unique;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_w = iem_gui_clip_size(w);
    vu_check_height(x, h);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    if(scale != 0)
	scale = 1;
    x->x_scale = scale;
    x->x_selected = 0;
    x->x_shiftdown = 0;
    x->x_peak = 0;
    x->x_rms = 0;
    x->x_fp = -101.0;
    x->x_fr = -101.0;
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft1"));
    x->x_out_rms = outlet_new(&x->x_obj, &s_float);
    x->x_out_peak = outlet_new(&x->x_obj, &s_float);
    return (x);
}

static void vu_free(t_vu *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void vu_setup(void)
{
    vu_class = class_new(gensym("vu"), (t_newmethod)vu_new, (t_method)vu_free,
			 sizeof(t_vu), 0, A_GIMME, 0);
    class_addbang(vu_class,vu_bang);
    class_addfloat(vu_class,vu_float);
    class_addmethod(vu_class, (t_method)vu_ft1, gensym("ft1"), A_FLOAT, 0);
    class_addlist(vu_class, vu_list);
    class_addmethod(vu_class, (t_method)vu_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_size, gensym("size"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_scale, gensym("scale"), A_DEFFLOAT, 0);
    class_addmethod(vu_class, (t_method)vu_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_color, gensym("color"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(vu_class, (t_method)vu_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(vu_class, (t_method)vu_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_label_font, gensym("label_font"), A_GIMME, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    vu_widgetbehavior.w_getrectfn =    vu_getrect;
    vu_widgetbehavior.w_displacefn =   vu_displace;
    vu_widgetbehavior.w_selectfn =     vu_select;
    vu_widgetbehavior.w_activatefn =   NULL;
    vu_widgetbehavior.w_deletefn =     vu_delete;
    vu_widgetbehavior.w_visfn =        vu_vis;
#if (PD_VERSION_MINOR > 31)
    vu_widgetbehavior.w_clickfn =      NULL;
    vu_widgetbehavior.w_propertiesfn = vu_properties;
#endif
    vu_widgetbehavior.w_savefn =       vu_save;
    class_setwidget(vu_class,&vu_widgetbehavior);
    class_sethelpsymbol(vu_class, gensym("vu"));
}

/* ------------------------ setup routine ------------------------- */

void g_7_guis_setup(void)
{
    bng_setup();
    hslider_setup();
    radiobut_setup();
    toggle_setup();
    my_canvas_setup();
    vslider_setup();
    vu_setup();
}
